
from os.path import basename, join, exists
import logging
import warnings
import numpy as np

LOG = logging.getLogger(__name__)


def read_marker_positions(species_marker_file):
    """Reads species marker file.

    Returns: Dict with start, end positions and len for each marker.
    """
    marker_positions = {}
    with open(species_marker_file, 'r') as marker_pos:
        header = next(marker_pos)
        for line in marker_pos:
            line = line.rstrip().split()
            marker = str(line[0]).strip()
            marker_start = int(line[1])
            marker_len = int(line[2])
            marker_end = marker_start + marker_len
            marker_positions[marker] = [marker_start, marker_end, marker_len]
    return marker_positions


def read_marker_list(marker_list_file):
    """Reads species marker list file.

    Returns: Set of marker names.
    """
    marker_list = set()
    with open(marker_list_file) as f:
        for l in f.readlines():
            marker_list.add(l.strip())
    return marker_list


def trunc_marker_ends(marker_pos, trunc_len):
    """Sets positions for truncation of marker ends.

    Returns: List of positions which are supposed to be
             truncated given `marker_pos` and `trunc_len`.
    """
    trunc_pos = set()
    for marker, (marker_start, marker_end, marker_len) in list(marker_pos.items()):

        # drop entirely if trunc_len * 2 longer than marker:
        if trunc_len * 2 > marker_len:
            LOG.info('Marker %s shorter than total `trunc_len`' % marker)
            trunc_pos = trunc_pos.union(list(range(marker_start, marker_end)))
            continue

        # set trunc positions
        trunc_pos_start = list(range(marker_start, marker_start + trunc_len))
        trunc_pos_end = list(range(marker_end - trunc_len, marker_end))
        trunc_pos = trunc_pos.union(trunc_pos_start, trunc_pos_end)
    return trunc_pos


def remove_marker_pos(marker_pos, remove_markers=None):
    """Sets positions for marker removal.

    Returns: Set of positions in `freqs` which are supposed to be
             removed given `remove_markers`.
    """
    remove_pos = set()
    for marker, (marker_start, marker_end, marker_len) in list(marker_pos.items()):
        if marker in remove_markers:
            remove_pos = remove_pos.union(
                list(range(marker_start, marker_end)))
    return remove_pos


def coverage(x, zero_nan=False):
    """Returns MxN numpy array with coverage for each sample."""
    cov = x.sum(axis=2)
    if zero_nan:
        cov[cov == 0] = np.nan
    return cov


def z_coverage(x, covered_pos_only=False):
    """Returns MxN numpy array with standardized coverage for each sample."""
    cov = coverage(x, zero_nan=covered_pos_only)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        zcov = (cov - np.nanmean(cov, axis=1)[:, np.newaxis]) / np.nanstd(
            cov, axis=1)[:, np.newaxis]
    return zcov


def species_min_samples(args, n_samples):
    if args['species_min_samples']:
        if n_samples < args['species_min_samples']:
            LOG.debug('Skipping %s since there are fewer than %s samples.' %
                      (args['species'], args['species_min_samples']))
            return False
    return True


def filter_freqs(args):

    # if exists, skip
    output_name = join(args['output_dir'], basename(args['input_file']))
    if exists(output_name):
        LOG.info('Skipping %s. Output file exists.' % args['species'])
        return True

    # load sample order
    with open(args['input_name'], 'r') as file:
        samples = file.read().strip().split('\n')

    # load sample selection
    if args['input_select'] is not None:
        with open(args['input_select'], 'r') as file:
            input_select = file.read().strip().split('\n')

    # skip if fewer than args['species_min_samples'] samples
    if not species_min_samples(args, len(samples)):
        return False

    # load freqs
    x = np.load(args['input_file'], allow_pickle=True)
    total_species_markers_size = x.shape[1]
    np.seterr(divide='ignore', invalid='ignore')
    removed_pos = set()

    # get marker metadata
    marker_metadata_file = join(args['marker_dir'],
                                '%s.positions.txt' % args['species'])
    marker_positions = read_marker_positions(marker_metadata_file)

    # 0 Remove Samples
    if args['input_select'] is not None:
        keep_samples = [n for n in range(
            0, len(samples)) if samples[n] in input_select]
        remove_n = len(samples) - len(keep_samples)
        samples = [samples[n] for n in keep_samples]
        LOG.info('Keeping %s out of %s selected sample(s). Removing: %s.' %
                 (len(keep_samples), len(input_select), remove_n))
        x = x[keep_samples, :, :]

    LOG.info('Filtering %s found in %s samples.' %
             (args['species'], len(samples)))

    # 1 Filter Markers

    # 1.1 Truncate ends
    if args['marker_trunc_len']:
        trunc_pos = trunc_marker_ends(marker_positions,
                                      args['marker_trunc_len'])

        LOG.info('Truncating marker ends at %s [%s%%] global positions.' %
                 (len(trunc_pos),
                  round(len(trunc_pos) / total_species_markers_size, 1)))

        # x has to be float (extract -> float not int)
        x[:, list(trunc_pos), :] = np.nan

        removed_pos = removed_pos.union(trunc_pos)

    # 1.2 Keep OR Remove markers
    if args['marker_keep'] or args['marker_remove']:

        if args['marker_keep']:
            # available marker set difference w/ markers
            keep_markers = read_marker_list(args['marker_keep']) & set(
                marker_positions.keys())
            rm_marker_set = set(keep_markers) ^ set(marker_positions.keys())

        elif args['marker_remove']:
            # available marker set intersect w/ markers
            rm_marker_set = read_marker_list(args['marker_remove']) & set(
                marker_positions.keys())

        remove_pos = remove_marker_pos(marker_positions,
                                       remove_markers=rm_marker_set)

        LOG.info('Zeroing %s markers at %s [%s%%] global positions.' %
                 (len(rm_marker_set), len(remove_pos),
                  round(len(remove_pos) / total_species_markers_size, 1)))
        x[:, list(remove_pos), :] = np.nan
        removed_pos = removed_pos.union(remove_pos)

    # 2 Filter Sample Variants [per Position]

    # 2.1 min-ncov
    if args['sample_var_min_n_vcov']:
        remove_pos = x < args['sample_var_min_n_vcov']
        x[remove_pos] = 0

    # 2.2 min-fcov
    if args['sample_var_min_f_vcov']:
        remove_pos = np.true_divide(
            x,
            coverage(x)[:, :, np.newaxis]) < args['sample_var_min_f_vcov']
        x[remove_pos] = 0

    # 3 Filter Positions [per Sample]

    # 3.1 min-vcov
    if args['sample_pos_min_n_vcov']:
        remove_pos = coverage(x) < args['sample_pos_min_n_vcov']
        x[remove_pos, :] = np.nan

    # 3.2 sd-vcov [of covered positions only]
    if args['sample_pos_min_sd_vcov']:
        remove_pos = abs(z_coverage(
            x, covered_pos_only=True)) > args['sample_pos_min_sd_vcov']
        x[remove_pos, :] = np.nan

    # 4 Filter Samples

    # 4.1 min-n or f horizontal coverage
    if args['samples_min_n_hcov']:
        covered = np.sum(coverage(x, zero_nan=True) > 0, axis=1)
        keep_samples = covered > args['samples_min_n_hcov']
        keep_samples_names = set(
            [samples[i] for i, b in enumerate(keep_samples) if b])
        remove_samples_names = keep_samples_names.symmetric_difference(samples)
        LOG.info('Removing %s sample(s) due to insufficient horizontal coverage: %s' %
                 (len(remove_samples_names), ', '.join(remove_samples_names)))
        samples = [s for s in samples if s not in remove_samples_names]
        x = x[keep_samples, :, :]

    if args['samples_min_f_hcov']:
        covered = np.sum(coverage(x, zero_nan=True) > 0, axis=1)
        keep_samples = np.true_divide(
            covered, x.shape[1]) > args['samples_min_f_hcov']
        keep_samples_names = set(
            [samples[i] for i, b in enumerate(keep_samples) if b])
        remove_samples_names = keep_samples_names.symmetric_difference(samples)
        LOG.info('Removing %s sample(s) due to insufficient horizontal coverage: %s' %
                 (len(remove_samples_names), ', '.join(remove_samples_names)))
        samples = [s for s in samples if s not in remove_samples_names]
        x = x[keep_samples, :, :]

    # 4.2 mean-vcov [of covered positions only]
    if args['samples_min_m_vcov']:
        keep_samples = np.nanmean(coverage(x, zero_nan=True),
                                  axis=1) > args['samples_min_m_vcov']
        keep_samples_names = set(
            [samples[i] for i, b in enumerate(keep_samples) if b])
        remove_samples_names = keep_samples_names.symmetric_difference(samples)
        LOG.info('Removing %s sample(s) due to insufficient vertical coverage: %s' %
                 (len(remove_samples_names), ', '.join(remove_samples_names)))
        samples = [s for s in samples if s not in remove_samples_names]
        x = x[keep_samples, :, :]

    # 4.4 Sample minimum
    if not species_min_samples(args, x.shape[0]):
        return False

    # 5 Filter Positions [Global]

    # 5.1 min [n or fraction] samples with coverage at position
    if args['global_pos_min_n_vcov'] or args['global_pos_min_f_vcov']:

        if args['global_pos_min_n_vcov']:
            remove_pos = np.sum(coverage(x) > 0,
                                axis=0) < args['global_pos_min_n_vcov']
        elif args['global_pos_min_f_vcov']:
            remove_pos = (np.sum(coverage(x) > 0, axis=0) /
                          x.shape[0]) < args['global_pos_min_f_vcov']

        LOG.info(
            'Zeroing %s global positions [%s%%] covered by too few samples.' %
            (sum(remove_pos),
             round(sum(remove_pos) / total_species_markers_size, 1)))
        x[:, remove_pos, :] = np.nan
        removed_pos = set([pos for pos, b in enumerate(remove_pos)
                           if b]).union(removed_pos)

    # 6 Filter Mono- and Polymorphic Positions
    if args['keep_poly'] or args['keep_mono']:

        if args['keep_poly']:
            remove_pos = ((x > 0).sum(axis=2) > 1).sum(axis=0) == 0
            LOG.info('Zeroing %s [%s%%] global monomorphic positions.' %
                     (sum(remove_pos),
                      round(sum(remove_pos) / total_species_markers_size, 1)))
        elif args['keep_mono']:
            remove_pos = ((x > 0).sum(axis=2) > 1).sum(axis=0) > 0
            LOG.info('Zeroing %s [%s%%] global polymorphic positions.' %
                     (sum(remove_pos),
                      round(sum(remove_pos) / total_species_markers_size, 1)))

        x[:, remove_pos, :] = np.nan
        removed_pos = set([pos for pos, b in enumerate(remove_pos)
                           if b]).union(removed_pos)

    # Delete positions from array
    if args['delete_pos']:
        x = np.delete(x, list(removed_pos), axis=1)

    LOG.info('Remaining positions: %s.' % x.shape[1])

    # 7 Clean up
    # 7.1 Filter samples with no coverage
    remove_samples = np.sum(coverage(x) > 0, axis=1) == 0
    samples = [s for i, s in enumerate(samples) if not remove_samples[i]]
    x = x[~remove_samples, :, :]
    LOG.info('Removing {} sample(s) with no coverage.'.format(sum(remove_samples)))

    # 7.2 Sample minimum
    if not species_min_samples(args, x.shape[0]):
        return False

    # Save retained samples to file
    samples_file = join(args['output_dir'], basename(args['input_name']))
    with open(samples_file, 'w') as file:
        txt = '\n'.join(samples)
        file.write(txt)

    # Save Array to file
    np.save(output_name, x, allow_pickle=True)



from os.path import exists
from os import makedirs
import logging
import glob

from samestr.summarize.read_metaphlan_data import merge_horizontal_metaphlan_tables, get_taxon_counts, get_taxon_cooccurrences, filter_metaphlan_profile, separate_metaphlan_taxonomy
from samestr.summarize.read_samestr_data import read_samestr_data, analyze_strain_events
import pandas as pd
import numpy as np

LOG = logging.getLogger(__name__)


def summarize(args):
    """
    Summarize SameStr results, including:
    - MetaPhlAn profiles
        - Taxonomic cooccurrences
    - SameStr results
        - Strain events
        - Strain cooccurrences
    """

    # create output directory
    if not exists(args['output_dir']):
        makedirs(args['output_dir'])

    # list metaphlan profiles
    mp_file_list = glob.glob(
        args['mp_profiles_dir'] + '/*' + args['mp_profiles_extension'])
    if not mp_file_list:
        raise ValueError('Could not find MetaPhlAn input files at: ' +
                         args['mp_profiles_dir'] + '*' + args['mp_profiles_extension'])

    # read metaphlan profiles and remove profile extension from file names
    LOG.info('Reading {} MetaPhlAn profiles from {}'.format(
        len(mp_file_list), args['mp_profiles_dir']))
    mp_profiles = merge_horizontal_metaphlan_tables(mp_file_list)
    mp_profiles.columns = mp_profiles.columns.str.replace(
        args['mp_profiles_extension'], '', regex=False)

    # get kingdom:species counts and cooccurrences
    mp_counts = get_taxon_counts(mp_profiles)
    mp_cooc = get_taxon_cooccurrences(mp_profiles)

    # filter only species level, expand taxonomy, and remove duplicates
    mp_species = filter_metaphlan_profile(mp_profiles, tax_level='species')
    mp_species = separate_metaphlan_taxonomy(mp_species, all_levels=True)
    mp_taxonomy = mp_species[['kingdom', 'phylum', 'class',
                              'order', 'family', 'genus', 'species']].drop_duplicates()

    # read samestr results
    LOG.info('Reading SameStr data from {}'.format(args['input_dir']))
    sstr_data = read_samestr_data(args['input_dir'])

    # add annotation for strain events and get possible and detected strain cooccurrences
    sstr_data, strain_cooc = analyze_strain_events(sstr_data,
                                                   overlap_threshold=args['aln_pair_min_overlap'],
                                                   similarity_threshold=args['aln_pair_min_similarity'])

    # merge all data, replace NaN with 0 only in the shared_strains and analyzed_strain column
    cooc_data = pd.merge(mp_cooc, strain_cooc, on=['row', 'col'], how='outer')
    cooc_data['analyzed_strain'] = cooc_data['analyzed_strain'].fillna(0)
    cooc_data['shared_strain'] = cooc_data['shared_strain'].fillna(0)
    cooc_data.iloc[:, 2:] = cooc_data.iloc[:, 2:].mask(
        cooc_data.iloc[:, 2:].isna(), other=np.nan).astype('Int64')

    # write output
    mp_taxonomy.to_csv(args['output_dir'] +
                       '/mp_taxonomy.tsv', sep='\t', index=False)
    mp_species.drop(['kingdom', 'phylum', 'class', 'order', 'family', 'genus'], axis=1).to_csv(
        args['output_dir'] + '/mp_species.tsv', sep='\t', index=False)
    mp_counts.to_csv(args['output_dir'] +
                     '/mp_counts.tsv', sep='\t', index=False)
    cooc_data.to_csv(args['output_dir'] +
                     '/sstr_cooccurrences.tsv', sep='\t', index=False)
    sstr_data.to_csv(args['output_dir'] +
                     '/sstr_strain_events.tsv', sep='\t', index=False)

    return args



from os.path import isfile

import logging

from samestr.utils import ooSubprocess

LOG = logging.getLogger(__name__)


def sam2bam(arg):
    """
        formats mapping files
        from sequence alignment format (sam)
        by filtering, converting, sorting, indexing
        to binary sequence alignment format (bam).
    """

    oosp = ooSubprocess.ooSubprocess(tmp_dir=arg['tmp_dir'])
    output_exists = isfile(arg['bam'])

    if not output_exists:
        error_pipe = None
        LOG.debug('Converting: %s' % (arg['sam']))
        # decompress
        p1 = oosp.chain('dump_file.py',
                        args=['--input_file', arg['sam']],
                        stderr=error_pipe)

        # filter len
        p2 = oosp.chain(
            'filter_sam.py',
            args=[
                str(int(arg['min_aln_identity'] * 100)),  # float
                str(arg['min_aln_len'])
            ],
            in_pipe=p1,
            stderr=error_pipe)

        # sam to bam
        p3 = oosp.chain('samtools',
                        args=['view', '-bS', '-'],
                        in_pipe=p2,
                        stderr=error_pipe)

        # sort
        oosp.chain('samtools',
                   args=['sort', '-', '-o', arg['bname']],
                   in_pipe=p3,
                   stderr=error_pipe,
                   out_fn=arg['bam'],
                   stop=True)

        # index
        oosp.ex('samtools', args=['index', arg['bam']], verbose=False)

    LOG.debug('Finished: %s' % arg['bam'])
    return arg



import pandas as pd
import numpy as np
from functools import reduce
from os.path import basename

import logging

LOG = logging.getLogger(__name__)


def get_metaphlan_taxonomy_levels():
    """
    Return a dict of the taxonomy levels that are used in MetaPhlAn.
    """
    return(
        {'kingdom':  {'prefix': 'k__', 'index': 0},
            'phylum':   {'prefix': 'p__', 'index': 1},
            'class':    {'prefix': 'c__', 'index': 2},
            'order':    {'prefix': 'o__', 'index': 3},
            'family':   {'prefix': 'f__', 'index': 4},
            'genus':    {'prefix': 'g__', 'index': 5},
            'species':  {'prefix': 's__', 'index': 6},
            'strain':   {'prefix': 't__', 'index': 7}
         }
    )


def read_metaphlan_profile(file_fn):
    """
    Read and parse a table to a dataframe, dropping the 'NCBI_tax_id' column and
    keeping only the 'clade_name' and 'relative_abundance' columns, and adding the
    'file_fn' column to the dataframe.
    """
    with open(file_fn, 'r') as f:
        lines = f.readlines()

    # Determine the format of the input file
    if lines[0].startswith('#SampleID\tMetaphlan_Analysis'):
        data_lines = lines[1:]
        col_names = ['clade_name', 'relative_abundance']
    else:
        data_lines = [line for line in lines if not line.startswith('#')]
        col_names = ['clade_name', 'NCBI_tax_id', 'relative_abundance']

    # Parse the data
    data = [line.strip().split('\t#')[0].split('\t') for line in data_lines]
    df = pd.DataFrame(data, columns=col_names)
    df['file_fn'] = basename(file_fn)

    # Drop the 'NCBI_tax_id' column if present
    if 'NCBI_tax_id' in df.columns:
        df = df.drop('NCBI_tax_id', axis=1)
    return df


def concat_vertical_metaphlan_tables(file_fns: list = None):
    """
    Read and concatenate (vertical) multiple metaphlan profile files into a single dataframe.
    """
    dfs = [read_metaphlan_profile(file_fn) for file_fn in file_fns]
    return pd.concat(dfs, ignore_index=True)


def merge_horizontal_metaphlan_tables(file_fns: list = None):
    """
    Read and merge (horizontal) multiple metaphlan profile files into a single dataframe,
    so that the column name for Abundance is the file name.
    """
    dfs = [read_metaphlan_profile(file_fn) for file_fn in file_fns]

    # set index
    dfs = [df.set_index('clade_name') for df in dfs]

    # rename columns to file name
    dfs = [df.rename(columns={'relative_abundance': df['file_fn'][0]}).drop(
        columns=['file_fn']) for df in dfs]

    # join the dataframes
    dfs = pd.concat(dfs, axis=1, join='outer').fillna(0)

    # bring the index to a 'clade_name' column
    dfs = dfs.reset_index().rename(columns={'index': 'clade_name'})

    return dfs


def filter_metaphlan_profile(df, tax_level):
    """
    Filter a metaphlan profile dataframe to only show rows where the lowest taxonomic
    level is the one specified.
    """
    # Get the index of the taxonomic level
    tax_levels = get_metaphlan_taxonomy_levels()
    level_index = tax_levels[tax_level]['index']

    # return rows where the count of pipes in the first column is level_index
    return df[df[df.columns[0]].apply(lambda x: x.count('|')) == level_index]


def separate_metaphlan_taxonomy(df, clade_column='clade_name', all_levels=False):
    """
    Split the 'clade_name' column into separate columns for each taxonomic level,
    using the taxonomic levels as column names and returning only the lowest taxonomic
    level if all_levels is False.
    """
    # get taxonomy level dict
    tax_levels = get_metaphlan_taxonomy_levels()

    # separate the clade column into separate columns within the df
    taxonomy_df = df[clade_column].str.split('\\|', expand=True)

    # name the columns with the taxonomic levels
    # and remove taxonomic prefixes from each column
    for idx in taxonomy_df.columns:
        for tax_level, tdata in tax_levels.items():
            if tdata['index'] == idx:
                taxonomy_df.rename(columns={idx: tax_level}, inplace=True)
                taxonomy_df[tax_level] = taxonomy_df[tax_level].str.replace(
                    tdata['prefix'], '')
    if not all_levels:
        # keep only the lowest taxonomic level
        taxonomy_df = taxonomy_df.iloc[:, -1:]

    # Concatenate the split dataframe with the original dataframe,
    # dropping the clade column
    separated_df = pd.concat(
        [taxonomy_df, df.drop(columns=[clade_column])], axis=1)
    return separated_df


def get_metaphlan_level(file_fns: list = None, tax_level=str, combine=['horizontal', 'vertical'], output_file=None, all_levels=True):
    """
    Read, concat, filter, and split multiple metaphlan profile files into a single dataframe,
    then save the result to a file if specified and return the result as a dataframe.
    """
    # Combine the metaphlan profile files
    if combine == 'horizontal':
        df = merge_horizontal_metaphlan_tables(file_fns)
    elif combine == 'vertical':
        df = concat_vertical_metaphlan_tables(file_fns)
    else:
        raise ValueError('Combine must be either "horizontal" or "vertical"')

    # Filter the dataframe to only show rows where the lowest taxonomic level is the one specified
    df = filter_metaphlan_profile(df, tax_level)

    # Split the 'clade_name' column into separate columns for each
    df = separate_metaphlan_taxonomy(
        df, clade_column='clade_name', all_levels=all_levels)

    # save to file if specified
    if output_file:
        df.to_csv(output_file, sep="\t", index=False)

    return df


def get_metaphlan_species_profile(ifn):
    """
    Read metaphlan_profile from `ifn`,
    return all species and relative abundances as dataframe
    """
    profile = get_metaphlan_level(file_fns=[ifn],
                                  tax_level='species',
                                  combine='horizontal',
                                  all_levels=False)
    return profile


def get_metaphlan_species_profile_dict(ifn):
    """
    Read metaphlan_profile from `ifn`,
    return all species and relative abundances as dict
    """
    profile = get_metaphlan_species_profile(ifn)
    return profile.set_index('species')[profile.columns[1]].to_dict()


def binarize_metaphlan_profile(df):
    """
    Take a metaphlan profile dataframe and return a binary dataframe
    """
    # index on first column
    df = df.set_index(df.columns[0])

    # convert to binary > numeric
    df = df.applymap(lambda x: 1 if x != 0 else 0)

    return df


def get_taxon_cooccurrence(wide_metaphlan_profile):
    """
    Take a horizontally merged, filtered, and tax separated metaphlan profile matrix
    and return a dataframe with the co-occurrence of clades 
    for each sample pair and taxonomic level
    this does not return self-co-occurrences and only one side of the matrix
    """
    # convert to binary > numeric, index on first column
    df = binarize_metaphlan_profile(wide_metaphlan_profile)

    # calculate co-occurrence
    cooc = np.dot(df.T, df)

    # convert to long format
    cooc_long = pd.DataFrame(cooc, columns=df.columns,
                             index=df.columns).stack().reset_index()

    # name the columns "row", "col", and "cooc"
    cooc_long.columns = ['row', 'col', 'cooc']

    # remove self-co-occurrences, one side of the matrix
    cooc_long = cooc_long[cooc_long['row'] < cooc_long['col']]

    return cooc_long


def get_taxon_cooccurrences(wide_metaphlan_profile):
    """
    Take a horizontally merged metaphlan profile matrix
    calculate the co-occurrence of clades for each taxonomic level
    and return a dataframe with co-occurrence counts for each sample pair and taxonomic level
    i.e. shared species, shared genera, shared families, etc.
    this does not return self-co-occurrences and only one side of the matrix
    """
    # get the taxonomic levels
    tax_levels = get_metaphlan_taxonomy_levels()
    coocs = []

    # filter the dataframe to only show rows where the lowest taxonomic level is the one specified
    for tax_level, tdata in tax_levels.items():

        if not tax_level == 'strain':
            # filter the dataframe to only show rows where the lowest taxonomic level is the one specified
            df = filter_metaphlan_profile(wide_metaphlan_profile, tax_level)
            df = separate_metaphlan_taxonomy(df, all_levels=False)
            cooc = get_taxon_cooccurrence(df)

            # renaming the cooc column to include the taxonomic level
            cooc.rename(columns={'cooc': 'shared_' + tax_level}, inplace=True)

            # get the cooccurrence
            coocs += [cooc]

    # merge the cooccurrence dataframes by row, col,
    # and fill in the missing values with 0
    cooc = reduce(lambda left, right: pd.merge(left, right, on=[
                  'row', 'col'], how='outer'), coocs).fillna(0)
    return cooc


def get_taxon_counts(wide_metaphlan_profile):
    """
    Take a horizontally merged metaphlan profile matrix
    sum the number of taxa for each sample and taxonomic level
    and return them in a dataframe
    """
    # get the taxonomic levels
    tax_levels = get_metaphlan_taxonomy_levels()
    counts = []

    # filter the dataframe to only show rows where the lowest taxonomic level is the one specified
    for tax_level, tdata in tax_levels.items():

        if not tax_level == 'strain':
            # filter the dataframe to only show rows where the lowest taxonomic level is the one specified
            df = filter_metaphlan_profile(wide_metaphlan_profile, tax_level)

            # get the number of taxa for each sample
            count = binarize_metaphlan_profile(df).sum(axis=0)
            count.rename('count_' + tax_level, inplace=True)
            counts += [count]

    # merge the cooccurrence dataframes by row, col,
    # and fill in the missing values with 0
    counts = pd.concat(counts, axis=1)

    # reset index and name col 'Name'
    counts.reset_index(inplace=True)
    counts.rename(columns={'index': 'Name'}, inplace=True)
    
    return counts



from os.path import isfile
from os import remove
import logging

from glob import glob
from samestr.utils import ooSubprocess

LOG = logging.getLogger(__name__)


def bam2freq(arg):
    """
        converts mapping files
        from binary sequence alignment (bam) format
        to SNV profiles (numpy arrays).
    """

    oosp = ooSubprocess.ooSubprocess(tmp_dir=arg['tmp_dir'])
    if not arg:
        LOG.error('Empty input')
        return False

    if not isfile(arg['kp']):
        LOG.debug('Piling: %s' % arg['bam'])
        oosp.ex(
            'kpileup.py',
            args=[
                arg['bname'],
                arg['bam'],
                arg['gene_file'],
                str(arg['min_base_qual']),
                str(arg['min_aln_qual']),
                str(arg['min_vcov'])
            ],
            out_fn=arg['kp'],
            verbose=False)

    if not glob('%s/*.np.cPickle' % arg['np']) or not glob(
            '%s/*.npy' % arg['np']):
        LOG.debug('Converting: %s' % arg['kp'])
        oosp.ex('kp2np.py',
                args=[
                    '--kp', arg['kp'], '--map', arg['contig_map'], '--sample',
                    arg['bname'], '--gene-file', arg['gene_file'],
                    '--output-dir', arg['np']
                ],
                verbose=False)
    else:
        LOG.warning('Skipping: %s. '
                    'Directory already contained np files: %s' %
                    (arg['bname'], arg['np']))
        
    # clean up
    if not arg['keep_tmp_files']:
        remove(arg['gene_file'])
        remove(arg['contig_map'])
        remove(arg['kp'])
        remove(arg['bam'])
        remove(arg['bam'] + '.bai')

    LOG.debug('Finished: %s' % arg['np'])
    return arg



#!/usr/bin/env python3
import sys
import subprocess
import argparse
import re
from collections import defaultdict

parser = argparse.ArgumentParser(
    description='This script is for parsing the BAM file and look for reads overlapping with the target genes and report the pileup.')
parser.add_argument('sample_id', help='sample ID')
parser.add_argument('bam_file', help='an aligned bam file')
parser.add_argument(
    'gene_file', help='a tab-delimited file of six columns in this order: contigId, geneId, begin, end, strand, DNA transcript seq. (Note: begin<end)')
parser.add_argument('min_bq', type=int,
                    help='the minimum base quality score of a sequenced base')
parser.add_argument('min_mq', type=int,
                    help='the minimum MQ mapping score of the aligned reads')
parser.add_argument('min_d', type=int, help='the minimum depth, an integer.')
args = parser.parse_args()

usage = f"""{sys.argv[0]} <sampleId> <bam file> <gene file> <minimum BQ> <minimum MQ> <minimum D> <maximum D>

{args.sample_id}: sample ID
{args.bam_file}: an aligned bam file
{args.gene_file}: a tab-delimited file of six columns in this order: contigId, geneId, begin, end, strand, DNA transcript seq. (Note: begin<end)
{args.min_bq}: the minimum base quality score of a sequenced base
{args.min_mq}: the minimum MQ mapping score of the aligned reads
{args.min_d}: the minimum depth, an integer.

This script is for parsing the BAM file and look for reads overlapping with the target genes and report the pileup.
Reads must pass the minimum MQ threshold.
For each position, an allele must have a depth >= <minimum D>.
The frequency of each allele is calculated after the filterings.

Note: the quality score must be Sanger Phred score (ascii 33).

Dependencies:
Use Samtools

CIGAR parser: the embedded cigar parser is not versatile, please review it to make sure that it is handling the cigar code appropriately.


The script generates a tab-delimited table directly to STDOUT.
Each row is a base of the queried genes.
A base could be reported repeatedly up to four times as four rows when polymorphisms are observed: A,T, G or C.
The columns are:
sample ID
contig ID
This Base Position
Gene ID
Ref allele of this Base
Condon Position (if coding gene)
Observed Consensus Allele of this Base (in this BAM)
Observed Allele
Coverage Depth of the Observed Allele
Allele Frequency of the Observed Allele

"""

def parse_gene(file):
    """
    Parse the input gene file

    Args:
        gene_file (str): the gene file name

    Returns:
        dict: a dictionary containing the gene name as key and the contig, start, end, strand, and sequence as values
    """
    data = {}
    with open(file, "r") as f:
        for line in f:
            line = line.strip()
            contig_id, gene_id, begin, end, strand, seq = line.split("\t")
            data[gene_id] = {
                'contig': contig_id,
                'begin': int(begin),
                'end': int(end),
                'strand': strand,
                'seq': seq
            }
    return data


def parse_bases(genes):
    """
    Go through each gene and add the nucleotide positions to a dictionary indexed by contigs

    Args:
        genes (dict): a dictionary containing gene data

    Returns:
        dict: a dictionary indexed by contigs and containing the gene name, reference nucleotide, and codon position as values
    """
    nuc = defaultdict(dict)
    for g, gene_data in genes.items():
        begin = gene_data['begin']
        end = gene_data['end']
        c = gene_data['contig']
        strand = gene_data['strand']
        temp = list(gene_data['seq'])

        for i in range(begin, end + 1):
            codon_pos = (i - begin + 1) % 3
            if strand == '-' and codon_pos != 2:
                codon_pos = 1 if codon_pos == 0 else 0
            codon_pos = 3 if codon_pos == 0 else codon_pos
            nuc[c][i] = f"{g}\t{temp[i-begin]}\t{codon_pos}"

    return nuc

def decode_cigar(cigar):
    """
    Decode the cigar string

    Args:
        cigar (str): the cigar string

    Returns:
        str: the decoded cigar string
    """
    cigar_parts = re.findall(r'(\d+)([MIDNSHP])', cigar)
    new_string = ''.join(c * int(n) for n, c in cigar_parts)
    return new_string


def convert_qual(qual_string):
    """
    Convert the quality string to a list of quality scores

    Args:
        qual_string (str): the quality string

    Returns:
        list: a list of quality scores
    """
    scores = [ord(q) - 33 for q in qual_string]
    return scores


def pileup(sample_id, bam_file, gene_file, min_bq, min_mq, min_depth):
    """
    Parse the BAM file and look for reads overlapping with the target genes and report the pileup

    Args:
        sample_id (str): the sample ID
        bam_file (str): the BAM file name
        gene_file (str): the gene file name
        min_bq (int): the minimum base quality score of a sequenced base
        min_mq (int): the minimum MQ mapping score of the aligned reads
        min_depth (int): the minimum depth, an integer.

    Returns:
        None
    """
    genes = parse_gene(gene_file)
    bases = parse_bases(genes)

    f_table = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))

    with subprocess.Popen(["samtools", "view", bam_file], stdout=subprocess.PIPE, universal_newlines=True) as bam_process:
        for line in bam_process.stdout:
            qname, flag, rname, begin, mapq, cigar, mrnm, mpos, isize, seq, qual, *info = line.strip().split('\t')

            if rname == "*" or int(mapq) < min_mq or rname not in bases:
                continue

            begin = int(begin)
            end = begin + len(seq) - 1
            qual_scores = convert_qual(qual)

            s = decode_cigar(cigar)
            b = list(seq)
            ci = list(s)

            new = []
            read_i = 0

            for cigar_i in range(len(ci)):
                base = "-"
                if ci[cigar_i] == "D":
                    base = "-"
                elif ci[cigar_i] == "H":
                    continue
                elif ci[cigar_i] in ["I", "S"]:
                    read_i += 1
                    continue
                elif qual_scores[read_i] < min_bq:
                    base = "-"
                    read_i += 1
                else:
                    base = b[read_i]
                    read_i += 1

                new.append(base)

            b = new
            for i in range(begin, begin + len(b)):
                nuc = b[i - begin]

                if bases[rname].get(i):
                    if nuc != "-":
                        f_table[rname][i][nuc] += 1

    for c in f_table:
        print(f"{c}===")
    print("Sample\tContig\tPosition\tGene\tRef\tCodon\tConsensus\tAllele\tCounts\tFrequency")

    for c in f_table:
        for pos in sorted(f_table[c]):
            total = 0
            major = ""

            for nuc in sorted(f_table[c][pos]):
                counts = f_table[c][pos][nuc]
                if counts < min_depth:
                    continue
                total += counts
                if major == "":
                    major = nuc
                elif f_table[c][pos][major] < counts:
                    major = nuc

            for nuc in sorted(f_table[c][pos]):
                counts = f_table[c][pos][nuc]
                if counts < min_depth:
                    continue
                percent = 100 * counts / total

                print(f"{sample_id}\t{c}\t{pos}\t{bases[c][pos]}\t{major}\t{nuc}\t{counts}\t{percent:.0f}")


def main():

    pileup(args.sample_id, 
           args.bam_file, 
           args.gene_file, 
           args.min_bq, 
           args.min_mq, 
           args.min_d)

if __name__ == "__main__":
    main()



class TaxClade:
    def __init__(self, name):
        self.children = {}
        self.name, self.father = name, None
        self.n_children = 0

    def update_child_count(self):
        self.n_children += 1
        cl = self.father
        while cl:
            cl.n_children += 1
            cl = cl.father

    def add_child(self, name):
        new_clade = TaxClade(name)
        self.children[name] = new_clade
        self.update_child_count()
        new_clade.father = self
        return new_clade

    def get_terminals(self):
        terms = []
        if not self.children:
            return [self]
        for c in list(self.children.values()):
            terms += c.get_terminals()
        return terms

    def get_full_name(self):
        fullname = [self.name]
        cl = self.father
        while cl:
            fullname = [cl.name] + fullname
            cl = cl.father
        return "|".join(fullname[1:])


class TaxTree:
    def __init__(self, mpa_pkl):
        self.root = TaxClade("root")
        self.all_clades = {}
        clades_txt = ((l.strip().split('|')[:-1])
                      for l, n in list(mpa_pkl['taxonomy'].items()))
        for clade in clades_txt:
            father = self.root
            for clade_lev in clade:
                if not clade_lev in father.children:
                    father.add_child(clade_lev)
                    self.all_clades[clade_lev] = father.children[clade_lev]
                father = father.children[clade_lev]

    def get_lineage(self, lineage):
        cl = self.root
        for clade in lineage.split('|'):
            cl = cl.children[clade]
        return cl



import numpy as np


def consensus(x):
    """
    Select dominant variants from each alignment.
    At ties, select randomly
    """

    # zeroing non-dominant variants, keeping ties of dominant variants
    d = x.copy()
    non_dominant = d < d.max(axis=2)[:, :, np.newaxis]
    d[non_dominant] = 0

    # from ties of dominant variants randomly select variant
    p_pos = np.array([0, 1, 2, 3])
    ties = d[(d > 0).sum(axis=2) > 1, :]
    for i, p in enumerate(ties):
        rand_select = np.array([0, 0, 0, 0])
        tie_idx = p_pos[(p > 0)]
        rand_select_idx = np.random.choice(tie_idx, 1)[0]
        rand_select[rand_select_idx] = p[rand_select_idx]
        ties[i] = rand_select
    d[(d > 0).sum(axis=2) > 1, :] = ties

    return d



#!/usr/bin/env python3
import subprocess
import os
import multiprocessing
from multiprocessing.pool import ThreadPool
import sys
from tempfile import NamedTemporaryFile
from . import which
import functools
import traceback
import logging

LOG = logging.getLogger(__name__)


class ooSubprocessException(Exception):
    pass


class ooSubprocess:
    def __init__(self, tmp_dir='tmp-ooSubprocess/'):
        self.chain_cmds = []
        self.tmp_dir = tmp_dir
        makedirs(tmp_dir)
        self.current_process = None

    def ex(self,
           prog,
           args=[],
           get_output=False,
           get_out_pipe=False,
           out_fn=None,
           in_pipe=None,
           verbose=True,
           **kwargs):

        try:
            if not which.is_exe(prog):
                raise ooSubprocessException(
                    "Error [ex]: cannot find the program {} in the executable path!".format(prog))

            if isinstance(args, str):
                args = args.split()

            if not isinstance(args, list):
                args = [args]

            cmd = [prog] + args
            print_cmd = 'ooSubprocess: ' + ' '.join(cmd)
            if verbose and out_fn and (not get_output):
                print_stderr(print_cmd + ' > ' + out_fn)
            elif verbose:
                print_stderr(print_cmd)

            if get_output:
                result = subprocess.check_output(cmd, stdin=in_pipe, **kwargs)
            elif get_out_pipe:
                tmp_file = NamedTemporaryFile(dir=self.tmp_dir)
                p = subprocess.Popen(cmd, stdin=in_pipe,
                                     stdout=tmp_file, **kwargs)
                self.current_process = p
                return_code = p.wait()
                tmp_file.seek(0)
                if in_pipe is not None:
                    in_pipe.close()

                result = tmp_file
            elif out_fn:
                with open(out_fn, 'w') as ofile:
                    result = subprocess.check_call(cmd,
                                                   stdin=in_pipe,
                                                   stdout=ofile,
                                                   **kwargs)
            else:
                result = subprocess.check_call(cmd, stdin=in_pipe, **kwargs)

            return result
        except KeyboardInterrupt:
            if self.current_process is not None:
                self.current_process.terminate()
            raise

    def chain(self,
              prog,
              args=[],
              stop=False,
              in_pipe=None,
              get_output=False,
              get_out_pipe=False,
              out_fn=None,
              verbose=True,
              **kwargs):

        try:
            if not which.is_exe(prog):
                raise ooSubprocessException(
                    "Error [chain]: cannot find the program {} in the executable path!".format(prog))

            if in_pipe is None and self.chain_cmds != []:
                raise ooSubprocessException(
                    "The pipeline was not stopped before creating a new one! In cache: {} in the executable path!".format(' | '.join(self.chain_cmds)))
            if out_fn and not stop:
                raise ooSubprocessException(
                    'out_fn (output_file_name) is only specified when stop = True!'
                )

            if isinstance(args, str):
                args = args.split()

            if not isinstance(args, list):
                args = [args]
            cmd = [prog] + args

            print_cmd = ' '.join(cmd)
            if out_fn and (not get_output):
                print_cmd += ' > ' + out_fn
            self.chain_cmds.append(print_cmd)

            if stop:
                if in_pipe is None:
                    raise ooSubprocessException(
                        'No input process to create a pipeline!')

                if verbose:
                    print_stderr('ooSubprocess: ' +
                                 ' | '.join(self.chain_cmds))

                self.chain_cmds = []
                if get_output:
                    result = subprocess.check_output(
                        cmd, stdin=in_pipe, **kwargs)
                elif get_out_pipe:
                    tmp_file = NamedTemporaryFile(dir=self.tmp_dir)
                    p = subprocess.Popen(cmd,
                                         stdin=in_pipe,
                                         stdout=tmp_file,
                                         **kwargs)
                    self.current_process = p
                    return_code = p.wait()

                    if return_code != 0:
                        raise ooSubprocessException(
                            'Failed when executing the command: {}\n'
                            'return code: {}'.format(' | '.join(self.chain_cmds), return_code))
                    tmp_file.seek(0)
                    if in_pipe is not None:
                        in_pipe.close()

                    result = tmp_file
                elif out_fn:
                    with (open(out_fn, 'w')) as ofile:
                        result = subprocess.check_call(cmd,
                                                       stdin=in_pipe,
                                                       stdout=ofile,
                                                       **kwargs)
                else:
                    result = subprocess.check_call(
                        cmd, stdin=in_pipe, **kwargs)
            else:
                tmp_file = NamedTemporaryFile(dir=self.tmp_dir)
                p = subprocess.Popen(cmd, stdin=in_pipe,
                                     stdout=tmp_file, **kwargs)
                self.current_process = p
                return_code = p.wait()
                if return_code != 0:
                    raise ooSubprocessException(
                        'Failed when executing the command: {}\n'
                        'return code: {}'.format(' | '.join(self.chain_cmds), return_code))
                tmp_file.seek(0)
                if in_pipe is not None:
                    in_pipe.close()

                result = tmp_file
            return result
        except KeyboardInterrupt:
            if self.current_process is not None:
                self.current_process.terminate()
            raise

    def ftmp(self, ifn):
        return os.path.join(self.tmp_dir, os.path.basename(ifn))


def fdir(dir, ifn):
    return os.path.join(dir, os.path.basename(ifn))


def makedirs(dir):
    if not os.path.exists(dir):
        try:
            os.makedirs(dir)
        except OSError as e:
            if e.errno != 17:
                raise
            pass
    elif not os.path.isdir(dir):
        raise ooSubprocessException(
            'Error: {} is not a directory!'.format(dir))


def replace_ext(ifn, old_ext, new_ext):
    if ifn[len(ifn) - len(old_ext):] != old_ext:

        new_ifn = ifn + new_ext
    else:
        new_ifn = ifn[:len(ifn) - len(old_ext)] + new_ext
    return new_ifn


def splitext(ifn):
    basename = os.path.basename(ifn)
    if ifn.endswith('.tar.bz2'):
        ext = '.tar.bz2'
    elif ifn.endswith('.tar.gz'):
        ext = '.tar.gz'
    else:
        ext = basename.split('.')[-1]
        if ext != basename:
            ext = '.' + ext
    base = basename[:-len(ext)]
    for t in ['.sam', '.fastq', '.fasta', '.fna']:
        if base.endswith(t):
            ext = t + ext
            base = base[:-len(t)]
    return base, ext


def trace_unhandled_exceptions(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except:
            raise Exception(traceback.format_exc())

    return wrapper


def parallelize(func, args, nprocs=1, use_threads=False):
    if nprocs > 1:
        if use_threads:
            pool = ThreadPool(nprocs)
        else:
            pool = multiprocessing.Pool(nprocs)
        results = pool.map(func, args)
        pool.close()
        pool.join()
    else:
        try:
            results = serialize(func, args)
        except Exception as e:
            results = serialize(e)
    return results


def parallelize_async(func, args, nprocs=1, use_threads=False):
    if nprocs > 1:
        if use_threads:
            pool = ThreadPool(nprocs)
        else:
            pool = multiprocessing.Pool(nprocs)
        app_results = []
        for a in args:
            app_results.append(pool.apply_async(func, [a]))
        pool.close()
        pool.join()
        results = [r.get() for r in app_results]
    else:
        results = serialize(func, args)
    return results


def serialize(func, args):
    results = []
    for arg in args:
        results.append(func(arg))
    return results


def print_stderr(*args):
    sys.stderr.write(' '.join(map(str, args)) + '\n')
    sys.stderr.flush()


def print_stdout(*args):
    sys.stdout.write(' '.join(map(str, args)) + '\n')
    sys.stdout.flush()



from os import listdir
import pandas as pd
import numpy as np
import logging

LOG = logging.getLogger(__name__)


def read_distmat(directory, species, suffix):
    """
    Function reads distance matrix from file
    returns distance matrix as data frame.
    """
    fn = f"{directory}/{species}{suffix}"
    distmat = pd.read_csv(fn, sep="\t", index_col=0)
    return distmat


def distmat_nonfinite_to_NA(distmat):
    """
    Function takes distance matrix as data frame as input
    transforms non-finite values to NA, 
    returns distance matrix as data frame.
    """
    distmat = distmat.mask(~np.isfinite(distmat))
    return distmat


def distmat_rm_upper_tri(distmat, rm_diag=False):
    """
    Function takes distance matrix as data frame as input
    and returns distance matrix as data frame without upper triangle.
    It also removes diagonal if rm_diag is True.
    """
    # Remove upper triangle
    distmat = distmat.where(np.tril(np.ones(distmat.shape)).astype(bool))

    # Remove diagonal
    if rm_diag:
        distmat = distmat.where(~np.eye(distmat.shape[0], dtype=bool))
    return distmat


def distmat_to_long(distmat, value_name):
    """
    Function takes distance matrix as data frame as input
    and returns a data frame in long format,
    containing all values from the distance matrix.
    """
    # Convert to long format
    distmat = distmat.stack().reset_index()
    distmat.columns = ['row', 'col', value_name]
    return distmat


def get_merged_distmats(directory, species):
    """
    Takes directory and species as input
    and returns a data.frame where file pairs for each species are merged.

    Requires:
    Input dir must contain samestr compare output files with
    extensions: .fraction.txt and .overlap.txt
    """
    # Read in the fraction and overlap files as dataframes
    distmat = read_distmat(directory, species, '.fraction.txt')
    overlap = read_distmat(directory, species, '.overlap.txt')

    # Convert non-finite values to NA
    distmat = distmat_nonfinite_to_NA(distmat)
    overlap = distmat_nonfinite_to_NA(overlap)

    # Convert wide to long format
    distmat_long = distmat_to_long(distmat, value_name='similarity')
    overlap_long = distmat_to_long(overlap, value_name='overlap')
    if distmat_long.shape[0] > 0 and overlap_long.shape[0] > 0:
        # Merge the two dataframes
        merged_distmat_overlap = pd.merge(distmat_long, overlap_long, on=[
                                          'row', 'col'], how='outer')

        # Add a column for species
        merged_distmat_overlap['species'] = species

        # remove self-matches, one side of the matrix
        merged_distmat_overlap = merged_distmat_overlap[merged_distmat_overlap['row']
                                                        < merged_distmat_overlap['col']]
        return merged_distmat_overlap


def read_samestr_data(sstr_dir):
    """
    Takes samestr compare output directory as input and runs get_merged_distmats
    """
    distance_species_list = [s.replace('.fraction.txt', '') for s in listdir(
        sstr_dir) if s.endswith('.fraction.txt')]
    overlap_species_list = [s.replace('.overlap.txt', '') for s in listdir(
        sstr_dir) if s.endswith('.overlap.txt')]
    species_intersect = set(overlap_species_list).intersection(
        set(distance_species_list))

    LOG.info(f'Merging matrix for {len(species_intersect)} species.')
    sstr_data = pd.DataFrame()
    for species in species_intersect:
        sstr_data = pd.concat(
            [sstr_data, get_merged_distmats(directory=sstr_dir, species=species)])

    return sstr_data


def analyze_strain_events(sstr_data, similarity_threshold=0.999, overlap_threshold=5000):
    """
    Takes samestr data as input and annotates it with analysis level, shared, and event.
    Also returns a data frame with the sum of strain co-occurrences
    across all species for each file pair, including zero counts.
    """
    sstr_data['analysis_level'] = sstr_data.apply(
        lambda x: 'strain-level' if x['overlap'] > overlap_threshold else 'species-level', axis=1)
    sstr_data['analyzed_strain'] = sstr_data['analysis_level'] == 'strain-level'
    sstr_data['shared_strain'] = (sstr_data['similarity'] >= similarity_threshold) & (
        sstr_data['analyzed_strain'] == True)
    sstr_data['event'] = sstr_data.apply(lambda x: 'shared_strain' if x['shared_strain']
                                         else 'other_strain' if x['overlap'] > overlap_threshold else 'same_species', axis=1)
    sstr_data = sstr_data[['row', 'col', 'species', 'similarity', 'overlap',
                           'analysis_level', 'analyzed_strain', 'shared_strain', 'event']]

    # get the sum of shared strains and analyzed strains for each file pair
    strain_cooccurrences = sstr_data.groupby(
        ['row', 'col'])[['shared_strain', 'analyzed_strain']].sum().reset_index()
        
    return sstr_data, strain_cooccurrences



#!/usr/bin/env python3
import os
import sys


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep) + sys.path:
            path = path.strip('"')
            exe_file = os.path.join(path, program)

            if is_exe(exe_file):
                return exe_file

    return None


def is_exe(program):
    return which(program) != None



from os.path import dirname, abspath, exists
from samestr.utils import ooSubprocess
from samestr.utils.utilities import list_str_all_endswith

import logging
LOG = logging.getLogger(__name__)


def set_output_structure(args):
    """Sets predefined output filenames.

    Additionally creates predefined output dirs if they don't exist.
    """

    # dest
    out_dir = abspath(args[0]['output_dir']) + '/'
    cmd = args[0]['command']

    ooSubprocess.makedirs(out_dir)

    if cmd == 'convert':

        for arg in args:
            n = arg['bname']

            # metaphlan
            arg['sam'] = arg['input_dir'] + n + '.sam.bz2'
            arg['bowtie2out'] = arg['input_dir'] + n + '.bowtie2out'

            # metaphlan profiles
            if not arg['mp_profiles_dir']:
                arg['mp_profiles_dir'] = arg['input_dir']
            else:
                arg['mp_profiles_dir'] = abspath(arg['mp_profiles_dir']) + '/'

            # file name
            arg['mp_profile'] = arg['mp_profiles_dir'] + n + arg[
                'mp_profiles_extension']

            # exists
            if not exists(arg['mp_profile']):
                LOG.error('MetaPhlAn file not found: %s' % arg['mp_profile'])
                exit(1)

            # output
            sample_dir = out_dir + n + '/'

            ## sam2bam
            arg['bam'] = sample_dir + n + '.bam'

            ## bam2freq
            arg['gene_file'] = sample_dir + n + '.gene_file.txt'
            arg['contig_map'] = sample_dir + n + '.contig_map.txt'
            arg['kp'] = sample_dir + n + '.kp.txt'
            arg['np'] = sample_dir

            # make sample dirs
            ooSubprocess.makedirs(sample_dir)

    return args


def spread_args_by_input_files(args):
    # spread args to list of args per sample/sample-pair
    spread_args = []
    for idx, (base_name,
              input_files) in enumerate(args['input_files'].items()):

        spread_args.append({})
        for arg in args:
            if not arg == 'input_files':
                spread_args[idx][arg] = args[arg]

        group_size = len(input_files)
        spread_args[idx]['bname'] = base_name
        spread_args[idx]['input_dir'] = dirname(abspath(input_files[0])) + '/'

        # for paired-end: sanity check for pair counts of two
        if args['input_sequence_type'] == 'paired':
            if group_size > 2:
                LOG.error('More than two samples (%s) found: %s [%s]' %
                          (group_size, base_name, ','.join(input_files)))
                exit(1)
            elif group_size < 2:
                LOG.error('Not all samples are paired: %s.'
                          'Check file extension after read index [%s]' %
                          (base_name, ','.join(input_files)))
                exit(1)
            else:
                spread_args[idx]['1%s' % args['input_extension']] = [
                    f for f in input_files if '1.fastq' in f
                ][0]
                spread_args[idx]['2%s' % args['input_extension']] = [
                    f for f in input_files if '2.fastq' in f
                ][0]

        # for single-end: sanity check 1 observation per base name
        else:
            if group_size > 1:
                LOG.error('More than one sample (%s) found: %s [%s]' %
                          (group_size, base_name, ','.join(input_files)))
                exit(1)
            else:
                spread_args[idx]['%s' %
                                 args['input_extension']] = input_files[0]

    return spread_args


def get_accepted_extension(file, accepted_extensions):
    e = '.' + '.'.join(file.rsplit('.', 1)[-1:])
    if e in accepted_extensions:
        return e
    else:
        e = '.' + '.'.join(file.rsplit('.', 2)[-2:])
        if e in accepted_extensions:
            return e
    return False


def get_uniform_extension(files, accepted_extensions):
    input_extension = get_accepted_extension(files[0], accepted_extensions)

    if not input_extension:
        LOG.error('Files must be supplied with accepted file extensions: %s' %
                  ', '.join(accepted_extensions))
        exit(1)

    if not list_str_all_endswith(files, input_extension):
        LOG.error('Not all files have the same file extension.')
        exit(1)

    return input_extension



#!/usr/bin/env python3
import re, sys

# Filter SAM file by percent identity and length

def decode_cigar(cigar):
    res = []
    hbeg = 0
    alnlen = 0
    hend = 0
    tlen = 0
    tabs = re.split('([a-zA-Z])', cigar)[:-1]
    for i in range(len(tabs)//2):
        ci = int(tabs[i*2])
        li = tabs[i*2 + 1]
        if li == 'H':
            if i == 0:
                hbeg += ci
            else:
                hend += ci
        if li == 'M':
            alnlen += ci
        if li != 'H' and li != 'D':
            tlen += ci
    return [hbeg, alnlen, hend, tlen]


def reverse_complement(seq):
    return seq.translate(str.maketrans('ACGTacgtNn', 'TGCAtgcaNn'))[::-1]


# read command line arguments and/or stdin
if (len(sys.argv[1:]) == 2):
    fhandle = sys.stdin.readlines()
    pctid = float(sys.argv[1])
    minlen = int(sys.argv[2])
else:
    with open(sys.argv[1], 'r') as f:
        fhandle = f.readlines()
        pctid = float(sys.argv[2])
        minlen = int(sys.argv[3])

# initialize variables
cquery = ''
cseq = ''
cqual = ''
cstrand = ''


# parse file
for line in fhandle:

    # print header
    if line.startswith('@'):
        print(line.rstrip())
        continue

    # get fields
    sline = line.rstrip().split('\t')
    query = sline[0]
    code = bin(int(sline[1]))[2:].zfill(12)
    strand = int(code[-5])
    ref = sline[2]
    cigar = sline[5]
    cigar = re.sub('H','S',cigar)
    sline[5] = cigar
    seq = sline[9]
    qual = sline[10]

    # skip empty hits
    if ref == '*' or cigar == '*':
        continue

    # make sure read is mapped
    if int(code[-3]) == 1:
        quit('ERROR 1: SAM file format')

    # calculate edit distance, total length
    [hbeg, alen, hend, tlen] = decode_cigar(cigar)
    mismatch = int(re.search('[NX]M:i:(\d+)', line).group(1))
    match = alen - mismatch

    # handle SAM asterisks
    if seq == '*' and qual == '*':
        # use last seq/qual
        if cquery != query:
            quit('ERROR 2: SAM file format')
    else:
        # update seq/qual
        cquery = query
        cseq = seq
        cqual = qual
        cstrand = strand

    # filter by percent identity
    if 1.*match/tlen < 1.*pctid/100.:
        continue

    # always set the seq/qual columns
    if strand == cstrand:
        sline[9] = cseq
        sline[10] = cqual
    else:
        sline[9] = reverse_complement(cseq)
        sline[10] = reverse_complement(cseq)

    # ensure that the cigar matches the sequence
    if tlen != (len(cseq) - hbeg - hend):
        quit('ERROR 3: SAM file format')

    # finally, print quality filtered line
    print('\t'.join(sline))



from datetime import datetime
from os.path import basename


def log_time(method):
    """
        Decorator for timing of function execution.
    """
    def timer(*args, **kw):
        t_start = datetime.now()
        result = method(*args, **kw)
        t_end = datetime.now()
        s = int((t_end - t_start).seconds)
        hours, remainder = divmod(s, 3600)
        minutes, seconds = divmod(remainder, 60)
        t_delta = '{:02}h:{:02}m:{:02}s'.format(int(hours), int(minutes),
                                                int(seconds))
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = t_delta
        else:
            if minutes < 1:
                print(('Finished "{}" in {:2}s'.format(method.__name__,
                                                       int(seconds))))
            elif hours < 1:
                print(('Finished "{}" in {:2}m {:2}s'.format(
                    method.__name__, int(minutes), int(seconds))))
            else:
                print(('Finished "{}" in {:2}h {:2}m {:2}s'.format(
                    method.__name__, int(hours), int(minutes), int(seconds))))

        return result

    return timer


def is_exe(name):
    """
        Check whether `name` is on PATH.
    """

    from distutils.spawn import find_executable
    return find_executable(name) is not None


def all_exe(exe_list):
    """
        Check whether each `exe` in `exe_list` is on PATH.
    """

    for exe in exe_list:
        if not is_exe(exe):
            raise Exception('Executable for %s not found' % exe)
        print(('Found in path: %s') % exe)

    return True


def list_str_all_endswith(string_list, ending):
    """
        Returns True if all `list` items share the same `ending`
    """
    if not all([item.endswith(ending) for item in string_list]):
        return False
    return True


def str_cut_endswith(string, ending):
    """
        Cuts `string` at the `ending` position
    """
    if string.endswith(ending):
        return string[:-len(ending)]
    return string


def list_str_cut_endswith(string_list, endings):
    """
        Returns list of strings with `endings` removed
    """
    if type(endings) is not list:
        endings = list(endings)

    cut_strings = []
    for s in string_list:
        cut = False
        for e in endings:
            if s.endswith(e):
                cut_strings.append(str_cut_endswith(s, e))
                cut = True
        if not cut:
            cut_strings.append(s)

    return cut_strings


def list_group_by_basename(names, cut_name_endings):
    cut_names = list_str_cut_endswith(names, cut_name_endings)
    base_names = {}
    for cut_name, input_file in zip(cut_names, names):
        cut_name = basename(cut_name)
        if cut_name in base_names:
            base_names[cut_name].append(input_file)
        else:
            base_names[cut_name] = [input_file]
    return base_names



#!/usr/bin/env python3
"""
    ref2freq generates multiple sequence alignments from MetaPhlAn
    marker sequences and supplied reference genomes [e.g. RefSeq, NCBI-Genome]
    which are further reduced to all positions in the initial
    MetaPhlAn marker positions and concatenated.

    Adapted StrainPhlAn's code [Truong et al.; Segata Lab]
    for adding references to MSA of sample consensus sequences
    to output to compatible numpy format
"""

from os import remove, makedirs
from os.path import join, isdir
from sys import exit

import random
from shutil import copyfile
from glob import glob
from re import sub

import bz2
import gzip
from collections import defaultdict
from tempfile import SpooledTemporaryFile, NamedTemporaryFile

import logging
import logging.config

import numpy as np

from Bio import SeqIO, Seq, SeqRecord, AlignIO
from Bio.Align import MultipleSeqAlignment

from samestr.utils import ooSubprocess
from samestr.utils.ooSubprocess import trace_unhandled_exceptions
from samestr.utils.utilities import all_exe

LOG = logging.getLogger(__name__)


def truncate_markers(args, reference_markers):
    """
        For each reference and marker in `reference_markers` dict,
        truncates marker by `marker_trunc_len` bases at beginning and end of the sequence
        and returns dict with truncated MSA
    """

    for reference in reference_markers:
        initial_markers_total_length = 0
        truncated_markers_total_length = 0
        remove_markers = []

        for marker in reference_markers[reference]:
            initial_markers_total_length += len(
                reference_markers[reference][marker]['seq'])

            # if marker_trunc_len, truncate
            if args['marker_trunc_len'] > 0:
                reference_markers[reference][marker]['seq'] = \
                    reference_markers[reference][marker]['seq'][args['marker_trunc_len']
                        :-args['marker_trunc_len']]
            truncated_markers_total_length += len(
                reference_markers[reference][marker]['seq'])

            # if empty, drop
            if not reference_markers[reference][marker]['seq']:
                remove_markers.append(marker)

        for marker in remove_markers:
            LOG.debug('Removing marker: %s' % marker)
            del reference_markers[reference][marker]

    return reference_markers


def align(args, marker_fn, aln_file):
    """
        Aligning `marker_fn` with `alignment program`
        returning alignment.
    """

    oosp = ooSubprocess.ooSubprocess(tmp_dir=args['tmp_dir'])

    # TODO: implement muscle 5 (double free corruption error if just one sequence)
    # return marker if marker has just one sequence
    # if len(list(SeqIO.parse(marker_fn, 'fasta'))) == 1:
    #     copyfile(marker_fn, aln_file)
    #     return

    if args['aln_program'] == 'muscle':
        alignment = oosp.ex('muscle',
                            # muscle 5 format, currently buggy
                            # args=['-quiet', '-align', marker_fn, '-output', aln_file],
                            # muscle v3.8.1551
                            args=['-quiet', '-in', marker_fn, '-out', aln_file],
                            get_output=True,  # wait for process to finish
                            verbose=False)

    elif args['aln_program'] == 'mafft':
        alignment = oosp.ex('mafft',
                            args=['--auto', '--quiet', marker_fn],
                            out_fn=aln_file,
                            verbose=False)
    return


def align_clean(args):
    """
        Creating clean alignment for marker of ooSubprocess instance using :align:
        returning multiple sequence alignment as dicts of numpy arrays:
            seq, DNA sequences: ['A', 'T', ..]
    """
    tmp_marker_file = NamedTemporaryFile(dir=args['tmp_dir'], mode='a+t', delete=False)
    tmp_marker_fn = tmp_marker_file.name

    # writing blast result of marker for each reference from dict to fasta file
    ref_count = 0
    for reference in list(args['reference_markers'].keys()):
        if args['species_marker'] in args['reference_markers'][reference]:
            ref_count += 1
            SeqIO.write(
                SeqRecord.SeqRecord(
                    id=reference,
                    description='',
                    seq=Seq.Seq(
                        args['reference_markers'][reference][args['species_marker']]['seq'])),
                tmp_marker_file, 'fasta')
    tmp_marker_file.close()

    # aligning sequences in fasta file
    tmp_alignment_file = NamedTemporaryFile(
        dir=args['tmp_dir'], mode='a+t', delete=not args['keep_tmp_files'])

    align(args, tmp_marker_fn, tmp_alignment_file.name)

    # adding alignment to out dicts
    seq = {}
    for rec in SeqIO.parse(tmp_alignment_file, 'fasta'):
        reference = rec.name
        seq[reference] = list(str(rec.seq))

    if args['alignment_fn']:
        copyfile(tmp_alignment_file.name, args['alignment_fn'])
    tmp_alignment_file.close()

    # delete temporary directory
    if not args['keep_tmp_files']:
        remove(tmp_marker_fn)

    if not seq:
        LOG.error('Fatal error in alignment step: %s' % species_marker)
        exit(1)

    return seq


def run_align_clean(args, reference_markers, species_markers_ordered):
    """
        Run :align_clean: on each marker in parallel with ooSubprocess
    """

    # parallelize align_clean
    args_list = []
    for i, _ in enumerate(species_markers_ordered):
        args_list.append({})
        args_list[i]['species_marker'] = species_markers_ordered[i]
        args_list[i]['species_marker_name'] = args['species_marker_name']
        args_list[i]['tmp_dir'] = args['tmp_dir']
        args_list[i]['output_dir'] = args['output_dir']
        args_list[i]['aln_program'] = args['aln_program']
        args_list[i]['reference_markers'] = reference_markers
        args_list[i]['keep_tmp_files'] = args['keep_tmp_files']
        if args['save_marker_aln']:
            formatted_marker_id = sub(r'[^a-zA-Z0-9 \n\.]', '_',
                                      species_markers_ordered[i])
            alignment_fn = join(args['output_dir'],
                                formatted_marker_id + '.marker_aligned.fasta')
            if args['marker_trunc_len'] > 0:
                alignment_fn = alignment_fn + '.marker_trunc_len_' + str(
                    args['marker_trunc_len'])
            args_list[i]['alignment_fn'] = alignment_fn
        else:
            args_list[i]['alignment_fn'] = False

    LOG.debug('Running align_clean for all markers')
    seqs = ooSubprocess.parallelize(align_clean,
                                    args_list,
                                    args['nprocs'],
                                    use_threads=False)

    if not seqs:
        LOG.error('No markers found!')
        exit(1)

    cat_seqs = defaultdict(list)

    # reformat dict to concatenate markers from seq[marker][ref] to seq[ref][marker]
    # remove positions where gaps were introduced into the species_marker
    # save positions where markers are concatenated, as well as marker lengths
    pos = 0
    marker_pos = []

    # seqs are ordered tuples
    for marker, _ in enumerate(seqs):
        if seqs[marker]:
            references_with_alignment = list(seqs[marker].keys())

            # index gaps in species_marker
            gap_positions = [
                idx for idx, n in enumerate(seqs[marker][
                    args['species_marker_name']]) if n == '-'
            ]
            marker_length = len(
                seqs[marker][args['species_marker_name']]) - len(gap_positions)

            if len(gap_positions):
                LOG.debug(
                    'Removing %s gap positions introduced during alignment: %s'
                    % (len(gap_positions), species_markers_ordered[marker]))

            # reduce marker to original mp_maker pos & length
            # add alignment seqs to reference dictionary
            for reference in list(reference_markers.keys()):
                if reference in references_with_alignment:
                    cat_seqs[reference] += [
                        n for idx, n in enumerate(seqs[marker][reference])
                        if idx not in gap_positions
                    ]

                # add empty arrays for references without marker in alignment
                else:
                    dummy_seqs = ['-' for p in range(marker_length)]
                    cat_seqs[reference] += dummy_seqs

            marker_pos.append(
                [species_markers_ordered[marker], pos, marker_length])
            pos += marker_length
        else:
            LOG.error('species_marker was not aligned: %s' %
                      species_markers_ordered[marker])
            exit(1)

    LOG.debug('Finished align_clean')
    return cat_seqs, marker_pos


def blast_markers_against_references(args, species_markers):
    """
        Using extract to find `species_markers` in BLASTdb generated from `input_files`
        returning search results in `reference_markers` as dict of markers:
            reference_markers[reference][marker_query]['seq']
    """

    oosp = ooSubprocess.ooSubprocess(tmp_dir=args['tmp_dir'])

    unique_markers = set(species_markers.keys())
    references = sorted(list(set(args['input_files'])))
    reference_markers = defaultdict(dict)
    contigs = defaultdict(dict)

    # read references from files
    p1 = SpooledTemporaryFile(mode='a+t', dir=args['tmp_dir'])
    for ref_fn in references:

        genome = ooSubprocess.splitext(ref_fn)[0]
        LOG.debug('Adding contigs for %s' % genome)
        if ref_fn.endswith('.bz2'):
            ref = bz2.open(ref_fn, 'rt')
        elif ref_fn.endswith('.gz'):
            ref = gzip.open(ref_fn, 'rt')
        elif ref_fn.endswith('.fna') or \
                ref_fn.endswith('.fa') or \
                ref_fn.endswith('.fasta'):
            ref = open(ref_fn, 'r')
        else:
            LOG.error('Unknown file type: %s '
                      'Fasta file should be raw [.fa, .fna, .fasta] '
                      'or zipped with bzip2 [.bz2] or gzip [.gz]' % ref_fn)
            exit(1)

        # check duplicate contigs
        unique_contigs_only = True
        for rec in SeqIO.parse(ref, 'fasta'):
            if rec.name in list(contigs.keys()):
                LOG.error('Contig %s is not unique' %
                          (rec.name.split('___')[-1]))
                unique_contigs_only = False
                break

        # only add references with before unseen contigs
        if not unique_contigs_only:
            LOG.warning('Skipping %s due to redundant contigs.' % genome)
            continue

        # load contigs from reference to dict, cat to tmp file
        else:
            LOG.debug('Contigs found for %s are unique' % genome)
            ref.seek(0)
            for rec in SeqIO.parse(ref, 'fasta'):
                contigs[rec.name]['seq'] = str(rec.seq)
                contigs[rec.name]['genome'] = genome
                SeqIO.write(rec, p1, 'fasta')

        ref.close()

    if not contigs:
        LOG.error('Failed to aggregate contigs.')
        exit(0)

    blastdb_prefix = oosp.ftmp('genome_extract_db_%s' % (random.random()))
    if glob('%s*' % blastdb_prefix):
        LOG.error('BLASTdb exists! Please remove it or rerun!')
        exit(1)

    # build blastdb from contigs of references in tmp file
    p1.seek(0)
    oosp.ex('makeblastdb',
            args=[
                '-dbtype', 'nucl', '-title', 'genome_db', '-out',
                blastdb_prefix
            ],
            in_pipe=p1,
            verbose=True)

    p1 = SpooledTemporaryFile(mode='a+t', dir=args['tmp_dir'])

    for unique_marker in unique_markers:
        SeqIO.write(species_markers[unique_marker], p1, 'fasta')

    # blast markers against reference contigs
    extract_args = [
        '-db', blastdb_prefix, '-outfmt', '6', '-evalue', '1e-10',
        '-word_size', '10', '-max_target_seqs', '1000000000'
    ]
    if args['nprocs'] > 1:
        extract_args += ['-num_threads', str(args['nprocs'])]

    p1.seek(0)
    extract_output = oosp.ex('blastn',
                             args=extract_args,
                             in_pipe=p1,
                             get_out_pipe=True,
                             verbose=True)

    # process extract output
    for line in extract_output:
        if line.strip() == '':
            break
        line = line.decode().strip().split()
        marker_query = line[0]
        target_contig = line[1]
        pstart = int(line[8]) - 1
        pend = int(line[9]) - 1

        reference = contigs[target_contig]['genome']

        # extract extract positions from reference contigs
        if marker_query not in reference_markers[reference]:
            reference_markers[reference][marker_query] = {}
            if pstart < pend:
                reference_markers[reference][marker_query]['seq'] = contigs[
                    target_contig]['seq'][pstart:pend + 1].upper()
            else:
                reference_markers[reference][marker_query]['seq'] = \
                    str(Seq.Seq(contigs[target_contig]['seq']
                        [pend:pstart+1]).reverse_complement()).upper()

    # remove database
    for fn in glob('%s*' % blastdb_prefix):
        remove(fn)

    LOG.debug('Number of reference genomes added: %d' % len(reference_markers))

    return reference_markers


def seqs2freqs(args, seqs, marker_pos):
    """
        Converts seqs to freqs and saves as numpy ndarrays,
        Reduces the alignment to positions only covered by the species_marker

        Info: Frequencies will always be 1 for references, since generated
        from BLAST and depth == 1. For mapped reads from samples,
        frequencies will be equivalent to the count of nucleotide observations
        at each position of the alignment e.g. [[14, 1, 0, 0], [0, 16, 0, 0],
        hence the sum of each vector equals coverage at given position.
    """

    # reduce to species_marker positions, duplicate, removing @align_clean to update positions
    species_marker_pos = np.array([
        pos for pos, n in enumerate(seqs[args['species_marker_name']])
        if n != '-'
    ])
    for reference in list(seqs.keys()):
        seqs[reference] = np.array(seqs[reference])[species_marker_pos]
    len_after_removal = len(seqs[args['species_marker_name']])

    species_marker_length_delta = len_after_removal - args[
        'species_markers_total_length']

    if species_marker_length_delta != 0:
        LOG.error('Difference to initial species_marker length: %s' %
                  str(species_marker_length_delta))
        exit(1)

    # nucleotide conversion arrays
    n_freq = np.array(
        [
            np.array([0, 0, 0, 0]),  # -
            np.array([0, 0, 0, 0]),  # N
            np.array([1, 0, 0, 0]),  # A
            np.array([0, 1, 0, 0]),  # C
            np.array([0, 0, 1, 0]),  # G
            np.array([0, 0, 0, 1])  # T
        ],
        dtype=float)
    acgt = '-NACGT'

    # convert seqs to freqs, set dimensions to [m=sample, n=width, k=4]
    freqs = defaultdict(list)
    for reference in list(seqs.keys()):
        freqs[reference] = np.array([
            n_freq[acgt.index(n)] if n in acgt else n_freq[acgt.index('-')]
            for n in seqs[reference]
        ])
        freqs[reference] = np.expand_dims(freqs[reference], axis=0)
        seqs[reference] = np.expand_dims(seqs[reference], axis=0)

    # makedirs
    if not isdir(args['output_dir']):
        makedirs(args['output_dir'])

    def _format_reference_name(name):
        # fasta, fna, fa, + gz
        if name.endswith(('.fa.gz', '.fasta.gz', '.fna.gz', '.fa.bz2',
                          '.fasta.bz2', '.fna.bz2')):
            return '.'.join(name.split('.')[:-2])
        elif name.endswith(('.fa', '.fna', '.fasta')):
            return '.'.join(name.split('.')[:-1])
        return name

    # save freqs array to file .npy
    output_base = args['output_dir'] + '/' + args['species']
    for reference in list(freqs.keys()):
        if reference.endswith('.markers'):
            continue

        fname = output_base + '.%s.npy' % _format_reference_name(reference)
        np.save(fname, freqs[reference], allow_pickle=True)

    # convert seqs to Bio SeqIO alignment object
    seqs_list = []
    name_order = []
    for reference in list(seqs.keys()):
        name_order.append(_format_reference_name(reference))
        seqs_list.append(
            SeqRecord.SeqRecord(id=_format_reference_name(reference),
                                description='',
                                seq=Seq.Seq(''.join(seqs[reference][0]))))
    seqs_msa = MultipleSeqAlignment(seqs_list)

    # write alignment fasta
    LOG.info('Saving alignment.')
    with open(output_base + '.msa.fa', 'w') as out:
        AlignIO.write(seqs_msa, out, 'fasta')

    return True


def check_dependencies(args):
    return all_exe([args['aln_program'], 'extract', 'makeblastdb'])


@trace_unhandled_exceptions
def ref2freq(args):
    # check_dependencies(args)

    # set species, tmp_dir, ofn, check output exists
    args['species_marker_name'] = args['species'] + '.markers'
    args['species_marker'] = args['marker_dir'] + '/' + args[
        'species_marker_name'] + '.fa'

    if args['tmp_dir'] == '':
        args['tmp_dir'] = args['output_dir']

    # load contigs from metaphlan marker
    species_markers = {}
    species_markers_ordered = []
    args['species_markers_total_length'] = 0
    for rec in SeqIO.parse(open(args['species_marker'], 'r'), 'fasta'):
        species_markers[rec.id] = rec
        species_markers_ordered.append(rec.id)
        args['species_markers_total_length'] += len(rec)

    # report basic info
    txt = 'Species: %s\nMarker File: %s\nMarkers: %s\nTotal Nucleotides: %s\nReferences: %s'
    LOG.info(
        txt %
        (args['species'], args['species_marker'], len(species_markers_ordered),
         args['species_markers_total_length'], ', '.join(args['input_files'])))

    # blast metaphlan markers against provided references
    reference_markers = blast_markers_against_references(args, species_markers)

    # add species_markers to extract output dict
    for species_marker in list(species_markers.keys()):
        reference_markers[args['species_marker_name']][species_marker] = {}
        reference_markers[
            args['species_marker_name']][species_marker]['seq'] = str(
                species_markers[species_marker].seq)

    # truncate marker ends
    if args['marker_trunc_len'] > 0:
        reference_markers = truncate_markers(args, reference_markers)

    # align markers, concatenate in order of species_markers
    seqs, marker_pos = run_align_clean(args, reference_markers,
                                       species_markers_ordered)

    # convert seqs to freqs, save msa, seqs, freqs, marker_pos
    seqs2freqs(args, seqs, marker_pos)




from os.path import basename, join, exists
import logging
import numpy as np

from Bio import Seq, SeqRecord, AlignIO
from Bio.Align import MultipleSeqAlignment

from samestr.filter import consensus

LOG = logging.getLogger(__name__)


def compare(args):

    # if exists, skip
    output_name = join(args['output_dir'], basename(args['input_file']))
    if exists(output_name):
        LOG.info('Skipping %s. Output file exists.' % args['species'])
        return True

    # load sample order
    with open(args['input_name'], 'r') as file:
        samples = file.read().strip().split('\n')

    # skip if fewer than args['samples_min_n'] samples
    if not len(samples) > 1:
        return False

    LOG.info('Comparing %s found in %s samples.' %
             (args['species'], len(samples)))

    # load freqs
    x = np.load(args['input_file'], allow_pickle=True)
    np.seterr(divide='ignore', invalid='ignore')

    # conversion arrays
    acgt = '-NACGT'
    n_freq = [
        [0, 0, 0, 0],  # -
        [0, 0, 0, 0],  # N
        [1, 0, 0, 0],  # A
        [0, 1, 0, 0],  # C
        [0, 0, 1, 0],  # G
        [0, 0, 0, 1]  # T
    ]

    # rename samples to samples.dom
    if args['dominant_variants'] or args['dominant_variants_added']:
        dom_samples = [s + '.dom' for s in samples]
        d = consensus(x)

    # analyze only dominant variants
    if args['dominant_variants']:
        x = d
        samples = dom_samples

    # add dominant variants separately as .dom samples
    elif args['dominant_variants_added']:
        x = np.append(x, d, axis=0)
        samples += dom_samples

    # set matrix colnames
    columns = 'Sample\t' + '\t'.join(samples)
    shortest_distance_matrix = [columns]
    shared_overlap_matrix = [columns]
    fraction_phenotype_matrix = [columns]

    # generate matrix: minimum variant similarity, overlap, fraction mvs
    for i in range(x.shape[0]):
        sample = samples[i]
        shortest_distance = (((x[i, :, :] > 0) *
                              (x > 0)).sum(axis=2) > 0).sum(axis=1)
        shared_overlap = ((x[i, :, :].sum(axis=1) > 0) *
                          (x.sum(axis=2) > 0)).sum(axis=1)
        fraction_phenotype = np.nan_to_num(shortest_distance / shared_overlap)
        shortest_distance_matrix.append(
            sample + '\t' +
            '\t'.join([str(n) for n in np.asarray(shortest_distance)]))
        shared_overlap_matrix.append(
            sample + '\t' +
            '\t'.join([str(n) for n in np.asarray(shared_overlap)]))
        fraction_phenotype_matrix.append(
            sample + '\t' +
            '\t'.join([str(n) for n in np.asarray(fraction_phenotype)]))

    # write matrix files
    with open('%s/%s.closest.txt' % (args['output_dir'], args['species']),
              'w') as out:
        out.write('\n'.join(shortest_distance_matrix))

    with open('%s/%s.overlap.txt' % (args['output_dir'], args['species']),
              'w') as out:
        out.write('\n'.join(shared_overlap_matrix))

    with open('%s/%s.fraction.txt' % (args['output_dir'], args['species']),
              'w') as out:
        out.write('\n'.join(fraction_phenotype_matrix))

    # output dominant variants as msa
    if 'dominant_variants_msa' in args and args['dominant_variants_msa']:

        if args['dominant_variants'] or args['dominant_variants_added']:
            samples = dom_samples
        else:
            d = consensus(x)

        # convert seqs to Bio SeqIO alignment object
        seqs_list = []
        for idx, freqs in enumerate((d > 0).astype(int).tolist()):
            seq = []
            for freq in freqs:
                try:
                    seq.append(acgt[n_freq.index(freq)])
                except ValueError:
                    max_idx = np.array([
                        i for i in range(0, len(freq)) if freq[i] == max(freq)
                    ])
                    random_max = np.random.choice(
                        max_idx, 1)[0] + 2  # for - and N in n_freq
                    seq.append(acgt[random_max])
            seqs_list.append(
                SeqRecord.SeqRecord(id=samples[idx],
                                    description=samples[idx],
                                    seq=Seq.Seq(''.join(seq))))
        seqs_msa = MultipleSeqAlignment(seqs_list)

        # write alignment fasta
        msa_filename = args['output_dir'] + '/' + args['species'] + '.msa.fa'
        with open(msa_filename, 'w') as out:
            AlignIO.write(seqs_msa, out, 'fasta')



#!/usr/bin/env python3
import argparse
import numpy as np
from os.path import isdir, basename
from os import makedirs

# Input arguments
# ---------------

parser = argparse.ArgumentParser()
parser.add_argument('--kp', help='Kpileup alignments (.kp.txt)')
parser.add_argument('--map', help='Map of genomes to contigs (tab-delimited)')
parser.add_argument('--sample', help='Sample Name', type=str, required=True)
parser.add_argument('--gene-file', help='kpileup gene file')
parser.add_argument('--output-dir', help='Output dir', default='./')
args = parser.parse_args()

# Read data
# ---------

# Numpy alignments
# x[contig] = numpy alignment
# y[genome] = concatenated numpy alignments
# Merge kpileups from multiple samples. Write dictionary of (M, N, 4) numpy arrays where:
# M = samples
# N = alignment sites
# 4 = nucleotides (ACGT)
# Entry (i,j,k) of this array corresponds to the count of nucleotide k at position j of sample i

# Initialize data
nts = 'ACGT'
M = 1

# Initialize numpy arrays for each contig
x = {}
for line in open(args.gene_file):
    line = line.rstrip().split()
    contig = line[0]
    beg = int(line[2])
    end = int(line[3])
    x[contig] = np.zeros([M, end, 4])

# Add kpileup results to numpy arrays
with open(args.kp, 'r') as f:
    for line in f.readlines():
        line = line.rstrip().split()
        if len(line) == 10 and line[0] != 'Sample':
            sample = line[0]
            i = 0
            contig = line[1]
            j = int(line[2])
            nt = line[7]
            k = nts.index(nt)
            count = int(line[8])
            x[contig][i, j - 1, k] = count

# Sample list
# M = [sample1]
M = np.array([args.sample])

# Contig map
# cmap[genome] = [contig1, contig2, ...]
cmap = {}
for line in open(args.map):
    line = line.strip().split()
    genome = line[0]
    contig = line[1]
    if genome not in cmap:
        cmap[genome] = []
    cmap[genome].append(contig)

# Create dir if not exists
if not isdir(args.output_dir):
    makedirs(args.output_dir)

# Mapping stats
cols = '\t'.join([
    'alignment', 'genome', 'mean_cov', 'median_cov', 'n_sites', 'n_gaps',
    'n_covered', 'n_mono', 'n_duo', 'n_tri', 'n_quat', 'n_poly', 'f_covered',
    'f_mono', 'f_duo', 'f_tri', 'f_quat', 'f_poly'
])
stats = [cols]

# Concatenate contigs
# -------------------
y = {}
for genome in cmap:
    contigs = cmap[genome]

    # Initialize array
    m = len(M)
    n = sum([np.shape(x[c])[1] for c in contigs])
    k = 4
    y[genome] = np.zeros([m, n, k])

    # Add alignment data
    beg = 0
    end = 0
    for contig in contigs:
        end += np.shape(x[contig])[1]
        y[genome][0, beg:end, :] = x[contig]
        beg = end

    np_filepath = '%s/%s.%s.npy' % (args.output_dir, genome, args.sample)
    np.save(np_filepath, y[genome], allow_pickle=True)

    species = y[genome]
    cov = species.sum(axis=2)

    # coverage [depth]
    mean_cov = round(np.mean(cov), 4)
    median_cov = round(np.median(cov), 4)

    # coverage [width]
    n_sites = species.shape[1]
    n_gaps = (cov == 0).sum()
    n_covered = n_sites - n_gaps

    # n of variant sites, monomorphic, .., polymorphic
    n_mono = ((species > 0).sum(axis=2) == 1).sum()
    n_duo = ((species > 0).sum(axis=2) == 2).sum()
    n_tri = ((species > 0).sum(axis=2) == 3).sum()
    n_quat = ((species > 0).sum(axis=2) == 4).sum()
    n_poly = ((species > 0).sum(axis=2) > 1).sum()

    # fraction of covered sites,
    # fraction of covered sites with variant, monomorphic, .., polymorphic
    if not n_covered == 0:
        f_covered = round(n_covered / n_sites, 4)
        f_mono = round(n_mono / n_covered, 4)
        f_duo = round(n_duo / n_covered, 4)
        f_tri = round(n_tri / n_covered, 4)
        f_quat = round(n_quat / n_covered, 4)
        f_poly = round(n_poly / n_covered, 4)
    else:
        f_covered, f_mono, f_duo, \
            f_tri, f_quat, f_poly = 0, 0, 0, 0, 0, 0

    stat = [
        basename(np_filepath), genome, mean_cov, median_cov, n_sites, n_gaps,
        n_covered, n_mono, n_duo, n_tri, n_quat, n_poly, f_covered, f_mono,
        f_duo, f_tri, f_quat, f_poly
    ]

    stat = [str(s) for s in stat]
    stats.append('\t'.join(stat))

with open('%s/%s.aln_stats.txt' % (args.output_dir, args.sample), 'w') as file:
    file.write('\n'.join(stats))



from os.path import isfile
import logging

from samestr.utils import ooSubprocess
from samestr.summarize.read_metaphlan_data import get_metaphlan_species_profile_dict

LOG = logging.getLogger(__name__)


def concatenate_gene_files(arg):
    """
        concatenates gene_files and contig_maps
        for all species of a given sample.
    """

    oosp = ooSubprocess.ooSubprocess(tmp_dir=arg['tmp_dir'])

    LOG.debug('Gathering: %s' % arg['mp_profile'])
    mp_species = get_metaphlan_species_profile_dict(arg['mp_profile'])
    if not mp_species:
        LOG.error('No species classified: %s' % arg['bname'])
        return False

    gene_files = []
    contig_maps = []

    for species, _ in sorted(iter(mp_species.items()),
                             key=lambda k_v: (k_v[1], k_v[0]),
                             reverse=True):
        gene_fname = '%s/%s.gene_file.txt' % (arg['marker_dir'], species)
        map_fname = '%s/%s.contig_map.txt' % (arg['marker_dir'], species)
        if isfile(gene_fname) and isfile(map_fname):
            gene_files.append(gene_fname)
            contig_maps.append(map_fname)

    if not isfile(arg['gene_file']):
        LOG.debug('Running: cat for %s gene_files' % len(gene_files))
        oosp.ex('cat', args=gene_files, out_fn=arg['gene_file'], verbose=False)

    if not isfile(arg['contig_map']) and len(contig_maps):
        LOG.debug('Running: cat for %s contig_maps' % len(contig_maps))
        oosp.ex('cat',
                args=contig_maps,
                out_fn=arg['contig_map'],
                verbose=False)

    LOG.debug('Finished: %s, %s' % (arg['contig_map'], arg['gene_file']))
    return arg



#!/usr/bin/env python3
# Original Author: Duy Tin Truong (duytin.truong@unitn.it)
# at CIBIO, University of Trento, Italy

import sys
import argparse as ap
import bz2
import gzip
import tarfile


def read_params():
    p = ap.ArgumentParser()
    p.add_argument('--input_file', required=True, default=None, type=str)
    return vars(p.parse_args())


def dump_file(ifn):
    file_ext = ''
    if ifn.endswith('.tar.bz2'):
        ifile = tarfile.open(ifn, 'r:bz2')
        file_ext = '.tar.bz2'
    elif ifn.endswith('.tar.gz'):
        ifile = tarfile.open(ifn, 'r:gz')
        file_ext = '.tar.gz'
    elif ifn.endswith('.bz2'):
        ifile = bz2.BZ2File(ifn, 'r')
        file_ext = '.bz2'
    elif ifn.endswith('.gz'):
        ifile = gzip.GzipFile(ifn, 'r')
        file_ext = '.gz'
    elif ifn.endswith('.fastq'):
        ifile = open(ifn, 'r')
        file_ext = '.fastq'
    elif ifn.endswith('.sam'):
        ifile = open(ifn, 'r')
        file_ext = '.sam'
    else:
        raise Exception('Unrecognized format! The format should be .bz2, .gz'
                        '.tar.bz2, .tar.gz, .sra, .sam.bz2, .sam, or .fastq\n')

    try:
        if file_ext in ['.tar.bz2', '.tar.gz']:
            for tar_info in ifile:
                ifile2 = ifile.extractfile(tar_info)
                if ifile2 != None:
                    for line in ifile2:
                        sys.stdout.write(line)
        else:
            for line in ifile:
                sys.stdout.write(line.decode('utf-8'))
    except:
        sys.stderr.write('Error while dumping file %s\n' % ifn)
        raise


if __name__ == "__main__":
    args = read_params()
    dump_file(args['input_file'])



from .summarize import summarize
from .read_metaphlan_data import merge_horizontal_metaphlan_tables, get_taxon_counts, get_taxon_cooccurrences, filter_metaphlan_profile, separate_metaphlan_taxonomy
from .read_samestr_data import read_samestr_data




from os.path import basename, join, exists
import logging
import warnings
import numpy as np
import pandas as pd
from scipy import stats
from samestr.filter import consensus

LOG = logging.getLogger(__name__)


def coverage(x):
    # calculate coverage for each sample
    # returns MxN numpy array
    return x.sum(axis=2)


def aln2stats(args):

    # if exists, skip
    output_name = join(args['output_dir'], basename(args['input_file']))
    if exists(output_name):
        LOG.info('Skipping %s. Output file exists.' % args['species'])
        return True

    # load sample order
    with open(args['input_name'], 'r') as file:
        samples = file.read().strip().split('\n')

    LOG.info('Gathering stats for %s found in %s samples.' %
             (args['species'], len(samples)))

    # load freqs
    x = np.load(args['input_file'], allow_pickle=True)
    total_species_markers_size = x.shape[1]
    np.seterr(divide='ignore', invalid='ignore')

    # conversion arrays
    acgt = '-NACGT'
    n_freq = [
        [0, 0, 0, 0],  # -
        [0, 0, 0, 0],  # N
        [1, 0, 0, 0],  # A
        [0, 1, 0, 0],  # C
        [0, 0, 1, 0],  # G
        [0, 0, 0, 1]  # T
    ]
    null_array = np.array([0, 0, 0, 0])

    # get dominant variants
    d = consensus(x)

    if args['dominant_variants']:
        # analyze only dominant variants
        x = d

    # suppress np warnings All-NaN Slice and Mean of empty slice
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', '', RuntimeWarning)

        # stats: vertical coverage
        cov = coverage(x)
        cov[cov == 0] = np.nan
        mean_cov = np.nanmean(cov, axis=1)
        mean_cov[np.where(np.isnan(mean_cov))] = 0
        median_cov = np.nanmedian(cov, axis=1)
        median_cov[np.where(np.isnan(median_cov))] = 0

        # stats: horizontal coverage
        n_sites = np.repeat(x.shape[1], x.shape[0])
        n_gaps = np.isnan(cov).sum(axis=1)
        n_covered = n_sites - n_gaps

        # stats: n of variant sites, monomorphic, .., polymorphic
        p_mono = ((x > 0).sum(axis=2) == 1)
        n_mono = p_mono.sum(axis=1)
        n_duo = ((x > 0).sum(axis=2) == 2).sum(axis=1)
        n_tri = ((x > 0).sum(axis=2) == 3).sum(axis=1)
        n_quat = ((x > 0).sum(axis=2) == 4).sum(axis=1)
        n_poly = ((x > 0).sum(axis=2) > 1).sum(axis=1)

        # polymorphic sites as per binomial cum. dist. func.
        illumina_error_rate = 0.3 / 100  # Q25+
        segata_error_rate = 1 / 100  # Q20
        p_value = 0.05
        n_binom = np.nansum(stats.binom.cdf(
            x.max(axis=2), cov, 1.0 - illumina_error_rate) < p_value,
            axis=1)
        f_binom = n_binom / n_covered
        n_binom_segata = np.nansum(stats.binom.cdf(
            x.max(axis=2), cov, 1.0 - segata_error_rate) < p_value,
            axis=1)
        f_binom_segata = n_binom_segata / n_covered

        # stats: fraction of covered sites,
        # stats: fraction of covered sites with variant, monomorphic, .., polymorphic
        f_covered = n_covered / n_sites
        f_mono = n_mono / n_covered
        f_duo = n_duo / n_covered
        f_tri = n_tri / n_covered
        f_quat = n_quat / n_covered
        f_poly = n_poly / n_covered

        # stats: vertical coverage of dominant variants
        # at all sites
        dom_cov = coverage(d)
        dom_cov[dom_cov == 0] = np.nan
        f_dom_cov = dom_cov / cov
        mean_dom_cov = np.nanmean(dom_cov, axis=1)
        mean_f_dom_cov = np.nanmean(f_dom_cov, axis=1)
        median_f_dom_cov = np.nanmedian(f_dom_cov, axis=1)

        # at polymorphic sites
        dom_cov[p_mono] = np.nan
        f_dom_cov = dom_cov / cov
        mean_dom_cov_polysites = np.nanmean(dom_cov, axis=1)
        median_dom_cov_polysites = np.nanmedian(dom_cov, axis=1)
        mean_f_dom_cov_polysites = np.nanmean(f_dom_cov, axis=1)
        median_f_dom_cov_polysites = np.nanmedian(f_dom_cov, axis=1)

        # mean coverage at polymorphic sites
        p_mono = np.where(np.isnan(dom_cov))
        cov[p_mono] = np.nan
        mean_cov_polysites = np.nanmean(cov, axis=1)

    # convert to pandas df
    df = pd.DataFrame(data=[
        np.array(samples), mean_cov, median_cov, n_sites, n_gaps, n_covered,
        n_mono, n_duo, n_tri, n_quat, n_poly, f_covered, f_mono, f_duo, f_tri,
        f_quat, f_poly, mean_dom_cov, mean_f_dom_cov, median_f_dom_cov,
        mean_dom_cov_polysites, median_dom_cov_polysites,
        mean_f_dom_cov_polysites, median_f_dom_cov_polysites,
        mean_cov_polysites, n_binom, f_binom, n_binom_segata, f_binom_segata
    ])
    df = df.T
    df.columns = [
        'Sample', 'mean_cov', 'median_cov', 'n_sites', 'n_gaps', 'n_covered',
        'n_mono', 'n_duo', 'n_tri', 'n_quat', 'n_poly', 'f_covered', 'f_mono',
        'f_duo', 'f_tri', 'f_quat', 'f_poly', 'mean_dom_cov', 'mean_f_dom_cov',
        'median_f_dom_cov', 'mean_dom_cov_polysites',
        'median_dom_cov_polysites', 'mean_f_dom_cov_polysites',
        'median_f_dom_cov_polysites', 'mean_cov_polysites', 'n_binom',
        'f_binom', 'n_binom_segata', 'f_binom_segata'
    ]

    # write df to file
    ofn = '%s/%s.aln_stats.txt' % (args['output_dir'], args['species'])
    df.to_csv(ofn, sep='\t', index_label=False, index=False)



from setuptools import setup, find_packages

with open('requirements.txt', 'rt') as requirements_file:
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]

setup(name='samestr',
      version='1.2023.04',
      description='SameStr identifies shared strains between pairs of '
                  'metagenomic samples based on the similarity of SNV profiles.',
      author='Daniel Podlesny',
      author_email='daniel.podlesny@embl.de',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      url='https://github.com/danielpodlesny/samestr/',
      license=open('LICENSE').read(),
      packages=find_packages(),
      package_data={'samestr': ['samestr', 'LICENSE']},
      install_requires=requirements,
      scripts=['samestr/samestr', 'samestr/convert/kpileup.py', 'samestr/convert/kp2np.py',
               'samestr/convert/dump_file.py', 'samestr/convert/filter_sam.py'],
      include_package_data=True)



from os.path import isfile
import logging
import numpy as np

LOG = logging.getLogger(__name__)


def merge_freqs(freqs, freq, freq_name):
    """Appends two numpy arrays."""

    if freqs.shape[1] == freq.shape[1]:
        _freqs = np.append(freqs, freq, axis=0)
        if _freqs.shape[0] == freqs.shape[0] + freq.shape[0]:
            return _freqs, True
        else:
            LOG.error('Could not add array: %s' % freq_name)
    else:
        LOG.error('Dimensions of arrays do not match: %s (%s|%s)' %
                  (freq_name, freqs.shape[1], freq.shape[1]))

    return freqs, False


def freq2freqs(args):
    """Merges nucleotide frequencies for individual species."""

    species_freqs = None
    samples = []

    # skip if species file exists
    output_file = '%s/%s' % (args['output_dir'], args['species'])
    if isfile('%s.npy' % output_file):
        LOG.info('Skipping: %s. Output file existed.' % args['species'])
        return True

    # iterate samples of species
    for sample, file_path in args['input_files']:

        # initialize species freqs arrays
        if species_freqs is None:
            species_freqs = np.load(file_path, allow_pickle=True)
            if isinstance(sample, list):
                samples += sample
            else:
                samples.append(sample)

            # if more than one sample, continue samples loop
            if len(args['input_files']) > 1:
                LOG.info('Merging %s inputs: %s' %
                         (len(args['input_files']), args['species']))
                continue

        # if more than one sample, add sample freq to freqs
        if len(args['input_files']) > 1:
            species_freqs, success = merge_freqs(
                species_freqs, np.load(file_path, allow_pickle=True), sample)
            if success:
                if isinstance(sample, list):
                    samples += sample
                else:
                    samples.append(sample)

    # per species, save freqs and names to file
    np.save(output_file, species_freqs, allow_pickle=True)

    with open(output_file + '.names.txt', 'w') as file:
        file.write('\n'.join(samples) + '\n')



from os.path import join
import bz2
import pickle as pickle
from Bio import SeqIO

import logging

from samestr.db import TaxClade, TaxTree


LOG = logging.getLogger(__name__)


def mp2db(input_args):
    """
    Function expands and generates a database of species markers from MetaPhlAn database
    that is required for SameStr.
    """

    # read mpa files into dict
    mpa_pkl = pickle.load(bz2.BZ2File(input_args['mpa_pkl']))
    if input_args['mpa_markers'].endswith('.bz2'):
        mpa_markers = SeqIO.to_dict(
            SeqIO.parse(bz2.open(input_args['mpa_markers'], 'rt'), 'fasta'))
    else:
        mpa_markers = SeqIO.to_dict(
            SeqIO.parse(input_args['mpa_markers'], 'fasta'))

    # set species markers
    all_species = set()
    all_markers = {}
    n_markers = 0

    # first, generate taxonomy
    tree = TaxTree(mpa_pkl)

    for marker in mpa_pkl['markers']:

        # retrieve species names using last clade that is not strain (sp. or higher)
        lineage = mpa_pkl['markers'][marker]['taxon'].split('|t__')[0]
        clade = lineage.split('|')[-1]

        if clade.startswith('s__'):
            species = clade.replace('s__', '')
        else:
            # mp marker tax. is not always the lowest available e.g.:
            # k__Bacteria|p__Actinobacteria|c__Actinobacteria|o__Bifidobacteriales|f__Bifidobacteriaceae|g__Gardnerella
            # but only species is s__Gardnerella_vaginalis
            node = tree.get_lineage(lineage)
            node_names = [n.name for n in node.get_terminals()]
            if len(node_names) == 1 and node_names[0].startswith('s__'):
                species = node_names[0].replace('s__', '')
            else:
                continue

        all_species.add(species)
        n_markers += 1

        # get only selected species
        if 'species' in input_args and input_args['species'] is not None:
            if species not in input_args['species']:
                continue

        if species not in all_markers:
            all_markers[species] = set()
        all_markers[species].add(marker)

    LOG.debug('MetaPhlAn db contains %s species, %s markers' %
              (len(all_species), n_markers))

    # report selected species
    if 'species' in input_args and input_args['species'] is not None:
        species_intersect = set(
            input_args['species']).intersection(all_species)
        species_diff = set(
            input_args['species']).difference(species_intersect)

        LOG.debug('Selected species found in the database: %s/%s' %
                  (len(species_intersect), len(input_args['species'])))
        if len(species_intersect) < len(input_args['species']):
            LOG.debug('Species not found in the database: %s' %
                      ', '.join(species_diff))
        all_species = species_intersect

    for species in all_species:

        # set output names
        output_base = join(input_args['output_dir'], species)
        marker_filename = output_base + '.markers.fa'
        pos_filename = output_base + '.positions.txt'
        gene_filename = output_base + '.gene_file.txt'
        cmap_filename = output_base + '.contig_map.txt'

        # check marker count
        species_markers = sorted(all_markers[species])
        n_markers = len(species_markers)
        LOG.debug('Markers found for %s: %s' % (species, n_markers))
        if not n_markers:
            continue

        total_marker_len = 0
        gene_data = []
        pos_data = ['%s\t%s\t%s' % ('contig', 'pos', 'len')]
        remove_markers = set()

        with open(marker_filename, 'w') as marker_file:

            for marker in species_markers:
                if marker in mpa_markers:
                    seq = mpa_markers[marker]
                    marker_len = len(seq)

                    contig_id = marker
                    gene_id = marker
                    strand = '+'
                    beg = 1
                    end = marker_len

                    pos_data += [
                        '%s\t%s\t%s' %
                        (marker, total_marker_len, marker_len)
                    ]
                    gene_data += ['%s\t%s\t%s\t%s\t%s\t%s' %
                                  (contig_id, gene_id, beg, end, strand, seq.seq)]
                    SeqIO.write(seq, marker_file, 'fasta')

                    total_marker_len += marker_len
                else:
                    remove_markers.add(marker)

        if remove_markers:
            LOG.debug('Dropping markers for %s: %s' %
                      (species, len(remove_markers)))
            species_markers = [
                m for m in species_markers if m not in remove_markers]

        with open(gene_filename, 'w') as gene_file, \
                open(cmap_filename, 'w') as cmap_file, \
                open(pos_filename, 'w') as pos_file:

            pos_file.write('\n'.join(pos_data) + '\n')
            gene_file.write('\n'.join(gene_data) + '\n')
            cmap_file.write('\n'.join(
                ['\t'.join([species, m]) for m in species_markers]) + '\n')
    return input_args
